import json
import math
import random
from typing import Sequence
import pytest
from pytest_httpserver import HTTPServer

from aleph_alpha_client import EmbeddingRequest
from aleph_alpha_client.aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.embedding import (
    BatchSemanticEmbeddingRequest,
    InstructableEmbeddingRequest,
    SemanticEmbeddingRequest,
    SemanticRepresentation,
    BatchSemanticEmbeddingResponse,
)
from aleph_alpha_client.prompt import Prompt

# AsyncClient


@pytest.mark.vcr
async def test_can_embed_with_async_client(async_client: AsyncClient, model_name: str):
    request = EmbeddingRequest(
        prompt=Prompt.from_text("abc"), layers=[-1], pooling=["mean"], tokens=True
    )

    response = await async_client.embed(request, model=model_name)
    assert response.model_version is not None
    assert response.embeddings and len(response.embeddings) == len(
        request.pooling
    ) * len(request.layers)
    assert response.tokens is not None
    assert response.num_tokens_prompt_total >= 1


@pytest.mark.vcr
async def test_can_semantic_embed_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = SemanticEmbeddingRequest(
        prompt=Prompt.from_text("hello"),
        representation=SemanticRepresentation.Symmetric,
        compress_to_size=128,
    )

    response = await async_client.semantic_embed(request, model=model_name)
    assert response.model_version is not None
    assert response.embedding
    assert len(response.embedding) == 128
    assert response.num_tokens_prompt_total >= 1


# for some reason this test does not work with vcr...
@pytest.mark.parametrize("num_prompts", [1, 100, 101])
@pytest.mark.parametrize("batch_size", [1, 32, 100])
async def test_batch_embed_semantic_with_async_client(
    async_client: AsyncClient, sync_client: Client, num_prompts: int, batch_size: int
):
    words = ["car", "elephant", "kitchen sink", "rubber", "sun"]
    r = random.Random(4082)
    prompts = list(
        [Prompt.from_text(words[r.randint(0, 4)]) for i in range(num_prompts)]
    )

    request = BatchSemanticEmbeddingRequest(
        prompts=prompts,
        representation=SemanticRepresentation.Symmetric,
        compress_to_size=128,
    )
    result = await async_client.batch_semantic_embed(
        request=request,
        num_concurrent_requests=10,
        batch_size=batch_size,
        model="luminous-base",
    )

    # We have no control over the exact tokenizer used in the backend, so we cannot know the exact
    # number of tokens. We do know, however, that there must be at least one token per prompt.
    assert result.num_tokens_prompt_total >= num_prompts

    assert len(result.embeddings) == num_prompts
    # To make sure that the ordering of responses is preserved,
    # we compare the returned embeddings with those of the sync_client's
    # sequential implementation
    embeddings_approximately_equal(
        result.embeddings, sync_client.batch_semantic_embed(request=request).embeddings
    )


@pytest.mark.vcr
@pytest.mark.parametrize("batch_size", [-1, 0, 101])
async def test_batch_embed_semantic_invalid_batch_sizes(
    async_client: AsyncClient, batch_size: int
):
    words = ["car", "elephant", "kitchen sink", "rubber", "sun"]
    request = BatchSemanticEmbeddingRequest(
        prompts=[Prompt.from_text(word) for word in words],
        representation=SemanticRepresentation.Symmetric,
    )

    with pytest.raises(ValueError):
        await async_client.batch_semantic_embed(request=request, batch_size=batch_size)


@pytest.mark.vcr
async def test_can_instructable_embed_with_async_client(
    async_client: AsyncClient,
):
    request = InstructableEmbeddingRequest(
        input=Prompt.from_text("hello"),
        instruction="Represent the text to query a database of salutations",
    )

    response = await async_client.instructable_embed(
        request, model="pharia-1-embedding-4608-control-with-256-embedding"
    )
    assert response.model_version is not None
    assert response.embedding
    assert len(response.embedding) == 4608
    assert response.num_tokens_prompt_total >= 1


def cosine_similarity(emb1: Sequence[float], emb2: Sequence[float]) -> float:
    """compute cosine similarity of v1 to v2: (v1 dot v2)/{||v1||*||v2||)"""
    sumxx, sumxy, sumyy = 0.0, 0.0, 0.0
    for i in range(len(emb1)):
        x = emb1[i]
        y = emb2[i]
        sumxx += x * x
        sumyy += y * y
        sumxy += x * y
    return sumxy / math.sqrt(sumxx * sumyy)


def embeddings_approximately_equal(a, b):
    assert all([cosine_similarity(v1, v2) > 0.99 for (v1, v2) in zip(a, b)])


async def test_model_name_gets_passed_along_for_async_client(httpserver: HTTPServer):
    request = BatchSemanticEmbeddingRequest(
        prompts=[Prompt.from_text("hello")],
        representation=SemanticRepresentation.Symmetric,
    )
    model_name = "test_model"
    expected_body = {
        **request.to_json(),
        "model": model_name,
    }
    httpserver.expect_ordered_request(
        "/batch_semantic_embed", method="POST", data=json.dumps(expected_body)
    ).respond_with_json(
        BatchSemanticEmbeddingResponse(
            model_version="1", embeddings=[], num_tokens_prompt_total=1
        ).to_json()
    )
    async_client = AsyncClient(token="", host=httpserver.url_for(""), total_retries=1)
    await async_client.batch_semantic_embed(request, model=model_name)


# Client


@pytest.mark.vcr
def test_embed(sync_client: Client, model_name: str):
    request = EmbeddingRequest(
        prompt=Prompt.from_text("hello"), layers=[0, -1], pooling=["mean", "max"]
    )

    result = sync_client.embed(request=request, model=model_name)

    assert result.model_version is not None
    assert result.embeddings and len(result.embeddings) == len(request.pooling) * len(
        request.layers
    )
    assert result.tokens is None
    assert result.num_tokens_prompt_total >= 1


@pytest.mark.vcr
def test_embedding_of_one_token_aggregates_identically(
    sync_client: Client, model_name: str
):
    request = EmbeddingRequest(
        prompt=Prompt.from_text(
            "hello"
        ),  # it is important for this test that we only embed one single token
        layers=[0, -1],
        pooling=["mean", "max"],
    )

    result = sync_client.embed(request, model=model_name)

    assert (
        result.embeddings
        and result.embeddings[("layer_0", "mean")]
        == result.embeddings[("layer_0", "max")]
    )


@pytest.mark.vcr
def test_embed_with_tokens(sync_client: Client, model_name: str):
    request = EmbeddingRequest(
        prompt=Prompt.from_text("abc"), layers=[-1], pooling=["mean"], tokens=True
    )

    result = sync_client.embed(request, model=model_name)

    assert result.model_version is not None
    assert result.embeddings and len(result.embeddings) == len(request.pooling) * len(
        request.layers
    )
    assert result.tokens is not None
    assert result.num_tokens_prompt_total >= 1


@pytest.mark.vcr
def test_embed_semantic(sync_client: Client):
    request = SemanticEmbeddingRequest(
        prompt=Prompt.from_text("hello"),
        representation=SemanticRepresentation.Symmetric,
        compress_to_size=128,
    )

    result = sync_client.semantic_embed(request=request, model="luminous-base")

    assert result.model_version is not None
    assert result.embedding
    assert len(result.embedding) == 128
    assert result.num_tokens_prompt_total >= 1


@pytest.mark.vcr
def test_embed_instructable(sync_client: Client):
    request = InstructableEmbeddingRequest(
        input=Prompt.from_text("hello"),
        instruction="Represent the text to query a database of salutations",
    )

    result = sync_client.instructable_embed(
        request=request, model="pharia-1-embedding-4608-control-with-256-embedding"
    )

    assert result.model_version is not None
    assert result.embedding
    assert len(result.embedding) == 4608
    assert result.num_tokens_prompt_total >= 1


@pytest.mark.vcr
@pytest.mark.parametrize("num_prompts", [1, 100, 101, 200])
def test_batch_embed_semantic(sync_client: Client, num_prompts: int):
    request = BatchSemanticEmbeddingRequest(
        prompts=[Prompt.from_text("hello") for _ in range(num_prompts)],
        representation=SemanticRepresentation.Symmetric,
        compress_to_size=128,
    )

    result = sync_client.batch_semantic_embed(request=request, model="luminous-base")
    assert len(result.embeddings) == num_prompts


def test_model_name_gets_passed_along_for_sync_client(httpserver: HTTPServer):
    request = BatchSemanticEmbeddingRequest(
        prompts=[Prompt.from_text("hello")],
        representation=SemanticRepresentation.Symmetric,
    )
    model_name = "test_model"
    expected_body = {**request.to_json(), "model": model_name}
    httpserver.expect_ordered_request(
        "/batch_semantic_embed", method="POST", data=json.dumps(expected_body)
    ).respond_with_json(
        BatchSemanticEmbeddingResponse(
            model_version="1", embeddings=[], num_tokens_prompt_total=1
        ).to_json()
    )
    sync_client = Client(token="", host=httpserver.url_for(""), total_retries=1)
    sync_client.batch_semantic_embed(request, model=model_name)



import importlib.metadata
from typing import Dict

MIN_API_VERSION = "1.19.0"


__version__ = importlib.metadata.version("aleph-alpha-client")


def user_agent_headers() -> Dict[str, str]:
    """User agent that should be send for specific versions of the SDK."""

    return {"User-Agent": "Aleph-Alpha-Python-Client-" + __version__}



import json

from packaging import version
from tokenizers import Tokenizer  # type: ignore
from types import TracebackType
from typing import (
    Any,
    AsyncGenerator,
    List,
    Mapping,
    Optional,
    Dict,
    Sequence,
    Type,
    Union,
)
import aiohttp
import asyncio
from aiohttp_retry import RetryClient, ExponentialRetry
import requests
from requests import Response
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict
from urllib3.util.retry import Retry
from tqdm.asyncio import tqdm

from aleph_alpha_client.explanation import (
    ExplanationRequest,
    ExplanationResponse,
)

from aleph_alpha_client.completion import (
    CompletionRequest,
    CompletionResponse,
    CompletionResponseStreamItem,
    stream_item_from_json,
)
from aleph_alpha_client.chat import (
    ChatRequest,
    ChatResponse,
    ChatStreamChunk,
    Usage,
    FinishReason,
    process_chat_stream,
    ToolCall,
)
from aleph_alpha_client.evaluation import EvaluationRequest, EvaluationResponse
from aleph_alpha_client.tokenization import TokenizationRequest, TokenizationResponse
from aleph_alpha_client.detokenization import (
    DetokenizationRequest,
    DetokenizationResponse,
)
from aleph_alpha_client.embedding import (
    BatchSemanticEmbeddingRequest,
    BatchSemanticEmbeddingResponse,
    EmbeddingRequest,
    EmbeddingResponse,
    EmbeddingVector,
    EmbeddingV2Request,
    EmbeddingV2Response,
    InstructableEmbeddingRequest,
    InstructableEmbeddingResponse,
    SemanticEmbeddingRequest,
    SemanticEmbeddingResponse,
)
from aleph_alpha_client.steering import (
    SteeringConceptCreationRequest,
    SteeringConceptCreationResponse,
)
from aleph_alpha_client.version import MIN_API_VERSION, user_agent_headers
from aleph_alpha_client.translation import TranslationRequest, TranslationResponse
from aleph_alpha_client.reranking import RerankRequest, RerankResponse

POOLING_OPTIONS = ["mean", "max", "last_token", "abs_max"]
RETRY_STATUS_CODES = frozenset({408, 429, 500, 502, 503, 504})
DEFAULT_REQUEST_TIMEOUT = 305


class QuotaError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class BusyError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _raise_for_status(status_code: int, text: str):
    if status_code >= 400:
        if status_code == 400:
            raise ValueError(status_code, text)
        elif status_code == 401:
            raise PermissionError(status_code, text)
        elif status_code == 402:
            raise QuotaError(status_code, text)
        elif status_code == 408:
            raise TimeoutError(status_code, text)
        elif status_code == 503:
            raise BusyError(status_code, text)
        else:
            raise RuntimeError(status_code, text)


def _check_api_version(version_str: str):
    api_ver = version.parse(MIN_API_VERSION)
    ver = version.parse(version_str)
    valid = api_ver.major == ver.major and api_ver <= ver
    if not valid:
        raise RuntimeError(
            f"The aleph alpha client requires at least api version {api_ver}, found version {ver}"
        )


AnyRequest = Union[
    CompletionRequest,
    ChatRequest,
    EmbeddingRequest,
    EmbeddingV2Request,
    EvaluationRequest,
    TokenizationRequest,
    DetokenizationRequest,
    SemanticEmbeddingRequest,
    InstructableEmbeddingRequest,
    BatchSemanticEmbeddingRequest,
    ExplanationRequest,
    ExplanationRequest,
    SteeringConceptCreationRequest,
    TranslationRequest,
    RerankRequest,
]


class Client:
    """
    Construct a client for synchronous requests given a user token

    Parameters:
        token (string, required):
            The API token that will be used for authentication.

        host (string, required):
            The hostname of the API host.

        hosting(string, optional, default None):
            Determines in which datacenters the request may be processed.
            You can either set the parameter to "aleph-alpha" or omit it (defaulting to None).

            Not setting this value, or setting it to None, gives us maximal flexibility in processing your request in our
            own datacenters and on servers hosted with other providers. Choose this option for maximal availability.

            Setting it to "aleph-alpha" allows us to only process the request in our own datacenters.
            Choose this option for maximal data privacy.

        request_timeout_seconds (int, optional, default 305):
            Client timeout that will be set for HTTP requests in the `requests` library's API calls.
            Server will close all requests after 300 seconds with an internal server error.

        total_retries(int, optional, default 8)
            The number of retries made in case requests fail with certain retryable status codes. If the last
            retry fails a corresponding exception is raised. Note, that between retries an exponential backoff
            is applied, starting with 0.5 s after the first retry and doubling for each retry made. So with the
            default setting of 8 retries a total wait time of 63.5 s is added between the retries.

        nice(bool, required, default False):
            Setting this to True, will signal to the API that you intend to be nice to other users
            by de-prioritizing your request below concurrent ones.

        verify_ssl(bool, optional, default True)
            Setting this to False will disable checking for SSL when doing requests.

        tags(Optional[Sequence[str]], optional, default None)
            Internal feature.

    Example usage:
        >>> request = CompletionRequest(prompt=Prompt.from_text(f"Request"), maximum_tokens=64)
        >>> client = Client(
                token=os.environ["TEST_TOKEN"],
                host=os.environ["TEST_API_URL"],
            )
        >>> response: CompletionResponse = client.complete(request, "pharia-1-llm-7b-control")
    """

    def __init__(
        self,
        token: str,
        host: str,
        hosting: Optional[str] = None,
        request_timeout_seconds: int = DEFAULT_REQUEST_TIMEOUT,
        total_retries: int = 8,
        nice: bool = False,
        verify_ssl=True,
        tags: Optional[Sequence[str]] = None,
        pool_size: int = 10,
    ) -> None:
        if host[-1] != "/":
            host += "/"
        self.host = host
        self.hosting = hosting
        self.request_timeout_seconds = request_timeout_seconds
        self.token = token
        self.nice = nice
        self.tags = tags

        retry_strategy = Retry(
            total=total_retries,
            backoff_factor=0.25,
            status_forcelist=RETRY_STATUS_CODES,
            allowed_methods=["POST", "GET"],
            raise_on_status=False,
        )
        adapter = HTTPAdapter(
            max_retries=retry_strategy,
            pool_connections=pool_size,
            pool_maxsize=pool_size,
        )
        self.session = requests.Session()
        self.session.verify = verify_ssl
        self.session.headers = CaseInsensitiveDict(
            {
                "Authorization": "Bearer " + self.token,
                **user_agent_headers(),
            }
        )
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    def validate_version(self) -> None:
        """Gets version of the AlephAlpha HTTP API."""
        _check_api_version(self.get_version())

    def get_version(self) -> str:
        """Gets version of the AlephAlpha HTTP API."""
        return self._get_request("version").text

    def _get_request(self, endpoint: str) -> Response:
        response = self.session.get(self.host + endpoint)
        if not response.ok:
            _raise_for_status(response.status_code, response.text)
        return response

    def _post_request(
        self,
        endpoint: str,
        request: AnyRequest,
        model: Optional[str] = None,
    ) -> Dict[str, Any]:
        json_body = self._build_json_body(request, model)

        query_params = self._build_query_parameters()

        response = self.session.post(
            self.host + endpoint,
            json=json_body,
            params=query_params,
            timeout=self.request_timeout_seconds,
        )
        if not response.ok:
            _raise_for_status(response.status_code, response.text)
        return response.json()

    def _build_query_parameters(self) -> Mapping[str, str]:
        return {
            # Cannot use str() here because we want lowercase true/false in query string.
            # Also do not want to send the nice flag with every request if it is false
            **({"nice": "true"} if self.nice else {}),
        }

    def _build_json_body(
        self, request: AnyRequest, model: Optional[str]
    ) -> Mapping[str, Any]:
        json_body = dict(request.to_json())

        if model is not None:
            json_body["model"] = model
        if self.hosting is not None:
            json_body["hosting"] = self.hosting
        if self.tags is not None:
            json_body["tags"] = self.tags
        return json_body

    def models(self) -> List[Mapping[str, Any]]:
        """
        Queries all models which are currently available.

        For documentation of the response, see
        https://docs.aleph-alpha.com/products/apis/pharia-inference/available-models/
        """
        response = self._get_request("models_available")
        return response.json()

    def complete(
        self,
        request: CompletionRequest,
        model: str,
    ) -> CompletionResponse:
        """Generates completions given a prompt.

        Parameters:
            request (CompletionRequest, required):
                Parameters for the requested completion.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # create a prompt
            >>> prompt = Prompt.from_text("An apple a day, ")
            >>>
            >>> # create a completion request
            >>> request = CompletionRequest(
                    prompt=prompt,
                    maximum_tokens=32,
                    stop_sequences=["###","\\n"],
                    temperature=0.12
                )
            >>>
            >>> # complete the prompt
            >>> result = client.complete(request, model=model_name)
        """
        response = self._post_request("complete", request, model)
        return CompletionResponse.from_json(response)

    def chat(
        self,
        request: ChatRequest,
        model: str,
    ) -> ChatResponse:
        """Chat with a model.

        Parameters:
            request (ChatRequest, required):
                Parameters for the requested chat.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> import os
            >>> from aleph_alpha_client import Client, ChatRequest, Message
            >>>
            >>> client = Client(token=os.environ["TOKEN"], host="https://inference-api.your.domain")
            >>> model = "llama-3.1-8b-instruct"
            >>> # create a chat request
            >>> request = ChatRequest(
                    messages=[Message(role="user", content="Hello, how are you?")],
                    model=model,
                )
            >>>
            >>> result = client.chat(request, model=model)
            >>> print(result.message)
        """
        response = self._post_request("chat/completions", request, model)
        return ChatResponse.from_json(response)

    def tokenize(
        self,
        request: TokenizationRequest,
        model: str,
    ) -> TokenizationResponse:
        """Tokenizes the given prompt for the given model.

        Parameters:
            request (TokenizationRequest, required):
                Parameters for the requested tokenization.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = TokenizationRequest(
                    prompt="hello", token_ids=True, tokens=True
                )
            >>> response = client.tokenize(request, model=model_name)
        """
        response = self._post_request(
            "tokenize",
            request,
            model,
        )
        return TokenizationResponse.from_json(response)

    def detokenize(
        self,
        request: DetokenizationRequest,
        model: str,
    ) -> DetokenizationResponse:
        """Detokenizes the given prompt for the given model.

        Parameters:
            request (DetokenizationRequest, required):
                Parameters for the requested detokenization.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = DetokenizationRequest(token_ids=[2, 3, 4])
            >>> response = client.detokenize(request, model=model_name)
        """
        response = self._post_request(
            "detokenize",
            request,
            model,
        )
        return DetokenizationResponse.from_json(response)

    def embed(
        self,
        request: EmbeddingRequest,
        model: str,
    ) -> EmbeddingResponse:
        """Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).

        Parameters:
            request (EmbeddingRequest, required):
                Parameters for the requested embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = EmbeddingRequest(prompt=Prompt.from_text(
                    "This is an example."), layers=[-1], pooling=["mean"]
                )
            >>> result = client.embed(request, model=model_name)
        """
        response = self._post_request(
            "embed",
            request,
            model,
        )
        return EmbeddingResponse.from_json(response)

    def embeddings(
        self,
        request: EmbeddingV2Request,
        model: str,
    ) -> EmbeddingV2Response:
        """Embeds a text and returns vectors that can be used for downstream tasks.
        This interface is compatible to the OpenAI /embeddings endpoint.

        Parameters:
            request (EmbeddingV2Request, required):
                Parameters for the requested embedding.

            model (string, required):
                Name of model to use.

        Examples:
            >>> request = EmbeddingV2Request(input="This is an example", dimensions=20)
            >>> result = client.embeddings(request, model=model_name)
        """
        response = self._post_request(
            "embeddings",
            request,
            model,
        )
        return EmbeddingV2Response.from_json(response)

    def semantic_embed(
        self,
        request: SemanticEmbeddingRequest,
        model: str,
    ) -> SemanticEmbeddingResponse:
        """Embeds a text and returns vectors that can be used for downstream tasks
        (e.g. semantic similarity) and models (e.g. classifiers).

        Parameters:
            request (SemanticEmbeddingRequest, required):
                Parameters for the requested semantic embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # function for symmetric embedding
            >>> def embed_symmetric(text: str):
                    # Create an embeddingrequest with the type set to symmetric
                    request = SemanticEmbeddingRequest(prompt=Prompt.from_text(
                        text), representation=SemanticRepresentation.Symmetric)
                    # create the embedding
                    result = client.semantic_embed(request, model=model_name)
                    return result.embedding
            >>>
            >>> # function to calculate similarity
            >>> def cosine_similarity(v1: Sequence[float], v2: Sequence[float]) -> float:
                    "compute cosine similarity of v1 to v2: (v1 dot v2)/{||v1||*||v2||)"
                    sumxx, sumxy, sumyy = 0, 0, 0
                    for i in range(len(v1)):
                        x = v1[i]; y = v2[i]
                        sumxx += x*x
                        sumyy += y*y
                        sumxy += x*y
                    return sumxy/math.sqrt(sumxx*sumyy)
            >>>
            >>> # define the texts
            >>> text_a = "The sun is shining"
            >>> text_b = "Il sole splende"
            >>>
            >>> # show the similarity
            >>> print(cosine_similarity(embed_symmetric(text_a), embed_symmetric(text_b)))
        """
        response = self._post_request(
            "semantic_embed",
            request,
            model,
        )
        return SemanticEmbeddingResponse.from_json(response)

    def batch_semantic_embed(
        self,
        request: BatchSemanticEmbeddingRequest,
        model: Optional[str] = None,
    ) -> BatchSemanticEmbeddingResponse:
        """Embeds a sequence of texts or images and returns vectors in the same order as they
        were provided. If more than 100 prompts are provided then this method will chunk them
        into batches of 100 prompts that will be sent to the API.

        Parameters:
            request (BatchSemanticEmbeddingRequest, required):
                Parameters for the requested semantic embeddings.

            model (string, optional, default None):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # function for symmetric embedding
            >>> def embed_symmetric(texts: Sequence[str]):
                    # Create an embeddingrequest with the type set to symmetric
                    request = BatchSemanticEmbeddingRequest(
                        prompts=[Prompt.from_text(text) for text in texts],
                        representation=SemanticRepresentation.Symmetric
                    )
                    # create the embedding
                    result = client.batch_semantic_embed(request, model=model_name)
                    return result.embedding
        """

        responses: List[EmbeddingVector] = []
        model_version = ""
        num_tokens_prompt_total = 0
        # The API currently only supports batch semantic embedding requests with up to 100
        # prompts per batch. As a convenience for users, this function chunks larger requests.
        for batch_request in _generate_semantic_embedding_batches(request):
            raw_response = self._post_request(
                "batch_semantic_embed",
                batch_request,
                model,
            )
            response = BatchSemanticEmbeddingResponse.from_json(raw_response)
            model_version = response.model_version
            responses.extend(response.embeddings)
            num_tokens_prompt_total += response.num_tokens_prompt_total

        return BatchSemanticEmbeddingResponse(
            model_version=model_version,
            embeddings=responses,
            num_tokens_prompt_total=num_tokens_prompt_total,
        )

    def instructable_embed(
        self,
        request: InstructableEmbeddingRequest,
        model: str,
    ) -> InstructableEmbeddingResponse:
        """Embeds a text and returns vectors that can be used for classification according to a given instruction.

        Parameters:
            request (InstructableEmbeddingRequest, required):
                Parameters for the requested instructable embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # function for salutation embedding
            >>> def embed_salutation(text: str):
                    # Create an embeddingrequest with a given instruction
                    request = InstructableEmbeddingRequest(
                        input=Prompt.from_text(text),
                        instruction="Represent the text to query a database of salutations"
                    )
                    # create the embedding
                    result = client.instructable_embed(request, model=model_name)
                    return result.embedding
            >>>
            >>> # function to calculate similarity
            >>> def cosine_similarity(v1: Sequence[float], v2: Sequence[float]) -> float:
                    "compute cosine similarity of v1 to v2: (v1 dot v2)/{||v1||*||v2||)"
                    sumxx, sumxy, sumyy = 0, 0, 0
                    for i in range(len(v1)):
                        x = v1[i]; y = v2[i]
                        sumxx += x*x
                        sumyy += y*y
                        sumxy += x*y
                    return sumxy/math.sqrt(sumxx*sumyy)
            >>>
            >>> # define the texts
            >>> text_a = "Hello"
            >>> text_b = "Good morning"
            >>>
            >>> # show the similarity
            >>> print(cosine_similarity(embed_salutation(text_a), embed_salutation(text_b)))
        """
        response = self._post_request(
            "instructable_embed",
            request,
            model,
        )
        return InstructableEmbeddingResponse.from_json(response)

    def evaluate(
        self,
        request: EvaluationRequest,
        model: str,
    ) -> EvaluationResponse:
        """Evaluates the model's likelihood to produce a completion given a prompt.

        Parameters:
            request (EvaluationRequest, required):
                Parameters for the requested evaluation.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = EvaluationRequest(
                    prompt=Prompt.from_text("hello"), completion_expected=" world"
                )
            >>> response = client.evaluate(request, model=model_name)
        """
        response = self._post_request(
            "evaluate",
            request,
            model,
        )
        return EvaluationResponse.from_json(response)

    def explain(
        self,
        request: ExplanationRequest,
        model: str,
    ) -> ExplanationResponse:
        """Better understand the source of a completion, specifically on how much each section of a
        prompt impacts each token of the completion.

        Parameters:
            request (ExplanationRequest, required):
                Parameters for the requested explanation.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = ExplanationRequest(
                    prompt=Prompt.from_text("Andreas likes"),
                    target=" pizza."
                )
            >>> response = client.explain(request, model="luminous-base")
        """
        response = self._post_request(
            "explain",
            request,
            model,
        )
        return ExplanationResponse.from_json(response)

    def create_steering_concept(
        self, request: SteeringConceptCreationRequest
    ) -> SteeringConceptCreationResponse:
        """Creates a steering concept.

        A steering concept consists of a list of steering examples. A steering
        example is a pair of a "negative" and a "positive" string, describing
        how you want to alter the model's output.

        This request will return a unique ID for the newly created steering
        concept that you can then use in completion and chat requests.

        Parameters:
            request (SteeringConceptCreationRequest, required)
                Parameters for the steering concepts to create.

        Examples:
            >>> request = SteeringConceptCreationRequest(
            >>>     examples=[
            >>>         SteeringPairedExample(
            >>>             negative="I appreciate your valuable feedback on this matter.",
            >>>             positive="Thanks for the real talk, fam.",
            >>>         ),
            >>>         SteeringPairedExample(
            >>>             negative="The financial projections indicate significant growth potential.",
            >>>             positive="Yo, these numbers are looking mad stacked!",
            >>>         ),
            >>>     ]
            >>> )
            >>> response = client.create_steering_concept(request)
        """
        response = self._post_request(
            "steering_concepts",
            request,
        )
        return SteeringConceptCreationResponse.from_json(response)

    def tokenizer(self, model: str) -> Tokenizer:
        """Returns a Tokenizer instance with the settings that were used to train the model.

        Examples:
            >>> tokenizer = client.tokenizer(model="luminous-base")
            >>> tokenized_prompt = tokenizer.encode("Hello world")
        """
        return Tokenizer.from_str(self._get_request(f"models/{model}/tokenizer").text)

    def translate(
        self,
        request: TranslationRequest,
    ) -> TranslationResponse:
        """Translates text from one language to another.

        Parameters:
            request (TranslationRequest, required):
                Parameters for the requested translation.

        Examples:
            >>> request = TranslationRequest(
                    model="pharia-1-mt-translation",
                    source="Hello, how are you?",
                    target_language="de"
                )
            >>> response = client.translate(request)
            >>> print(response.translation)
        """
        response = self._post_request("translate", request)
        return TranslationResponse.from_json(response)

    def rerank(
        self,
        request: RerankRequest,
        model: str,
    ) -> RerankResponse:
        """Reranks documents against a query.

        This endpoint takes in a query and a list of documents and produces an array
        with each document assigned a relevance score.

        Parameters:
            request (RerankRequest, required):
                Parameters for the requested reranking.

            model (string, required):
                Name of the model to use for reranking.

        Examples:
            >>> request = RerankRequest(
                    query="What is the capital of France?",
                    documents=[
                        "The capital of Brazil is Brasilia.",
                        "The capital of France is Paris.",
                        "Horses and cows are both animals.",
                    ],
                    top_n=2,
                )
            >>> response = client.rerank(request, model="your-reranker-model")
            >>> for result in response.results:
            >>>     print(f"Document {result.index}: {result.relevance_score}")
        """
        response = self._post_request("rerank", request, model)
        return RerankResponse.from_json(response)


class AsyncClient:
    """
    Construct a context object for asynchronous requests given a user token

    Parameters:
        token (string, required):
            The API token that will be used for authentication.

        host (string, required):
            The hostname of the API host.

        hosting(string, optional, default None):
            Determines in which datacenters the request may be processed.
            You can either set the parameter to "aleph-alpha" or omit it (defaulting to None).

            Not setting this value, or setting it to None, gives us maximal flexibility in processing your request in our
            own datacenters and on servers hosted with other providers. Choose this option for maximal availability.

            Setting it to "aleph-alpha" allows us to only process the request in our own datacenters.
            Choose this option for maximal data privacy.

        request_timeout_seconds (int, optional, default 305):
            Client timeout that will be set for HTTP requests in the `aiohttp` library's API calls.
            Server will close all requests after 300 seconds with an internal server error.

        total_retries(int, optional, default 8)
            The number of retries made in case requests fail with certain retryable status codes. If the last
            retry fails a corresponding exception is raised. Note, that between retries an exponential backoff
            is applied, starting with 0.25 s after the first request and doubling for each retry made. So with the
            default setting of 8 retries a total wait time of 63.75 s is added between the retries.

        nice(bool, required, default False):
            Setting this to True, will signal to the API that you intend to be nice to other users
            by de-prioritizing your request below concurrent ones.

        verify_ssl(bool, optional, default True)
            Setting this to False will disable checking for SSL when doing requests.

        tags(Optional[Sequence[str]], optional, default None)
            Internal feature.

    Example usage:
        >>> request = CompletionRequest(prompt=Prompt.from_text(f"Request"), maximum_tokens=64)
        >>> async with AsyncClient(
                token=os.environ["TEST_TOKEN"],
                host=os.environ["TEST_API_URL"],
            ) as client:
                response: CompletionResponse = await client.complete(request, "pharia-1-llm-7b-control")
    """

    def __init__(
        self,
        token: str,
        host: str,
        limit: int = 100,
        hosting: Optional[str] = None,
        request_timeout_seconds: int = DEFAULT_REQUEST_TIMEOUT,
        total_retries: int = 8,
        nice: bool = False,
        verify_ssl=True,
        tags: Optional[Sequence[str]] = None,
    ) -> None:
        if host[-1] != "/":
            host += "/"
        self.host = host
        self.hosting = hosting
        self.request_timeout_seconds = request_timeout_seconds
        self.token = token
        self.nice = nice
        self.tags = tags

        retry_options = ExponentialRetry(
            attempts=total_retries + 1,
            exceptions={aiohttp.ClientConnectionError},
            start_timeout=0.25,
            statuses=set(RETRY_STATUS_CODES),
        )
        connector = aiohttp.TCPConnector(verify_ssl=verify_ssl, limit=limit)
        self.session = RetryClient(
            trust_env=True,  # same behaviour as requests/(Sync)Client wrt. http_proxy
            raise_for_status=False,
            retry_options=retry_options,
            timeout=aiohttp.ClientTimeout(self.request_timeout_seconds),
            headers={
                "Authorization": "Bearer " + self.token,
                **user_agent_headers(),
            },
            connector=connector,
        )

    async def close(self):
        """Needs to be called at end of lifetime if the AsyncClient object is not used as a context manager."""
        await self.session.close()

    def __enter__(self) -> None:
        raise TypeError("Use async with instead")

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        # __exit__ should exist in pair with __enter__ but never executed
        pass  # pragma: no cover

    async def __aenter__(self) -> "AsyncClient":
        await self.session.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        await self.session.__aexit__(exc_type=exc_type, exc_val=exc_val, exc_tb=exc_tb)

    async def validate_version(self) -> None:
        _check_api_version(await self.get_version())

    async def get_version(self) -> str:
        """Gets version of the AlephAlpha HTTP API."""
        return await self._get_request_text("version")

    async def _get_request_text(self, endpoint: str) -> str:
        async with self.session.get(
            self.host + endpoint,
        ) as response:
            if not response.ok:
                _raise_for_status(response.status, await response.text())
            return await response.text()

    async def _get_request_json(
        self, endpoint: str
    ) -> Union[List[Mapping[str, Any]], Mapping[str, Any]]:
        async with self.session.get(
            self.host + endpoint,
        ) as response:
            if not response.ok:
                _raise_for_status(response.status, await response.text())
            return await response.json()

    async def _post_request(
        self,
        endpoint: str,
        request: AnyRequest,
        model: Optional[str] = None,
    ) -> Dict[str, Any]:
        json_body = self._build_json_body(request, model)

        query_params = self._build_query_parameters()
        async with self.session.post(
            self.host + endpoint, json=json_body, params=query_params
        ) as response:
            if not response.ok:
                _raise_for_status(response.status, await response.text())
            return await response.json()

    SSE_DATA_PREFIX = "data: "

    async def _post_request_with_streaming(
        self,
        endpoint: str,
        request: AnyRequest,
        model: Optional[str] = None,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        json_body = self._build_json_body(request, model)
        json_body["stream"] = True

        query_params = self._build_query_parameters()

        async with self.session.post(
            self.host + endpoint, json=json_body, params=query_params
        ) as response:
            if not response.ok:
                _raise_for_status(response.status, await response.text())

            async for stream_item in response.content:
                stream_item_as_str = stream_item.decode().strip()

                if not stream_item_as_str:
                    continue

                if not stream_item_as_str.startswith(self.SSE_DATA_PREFIX):
                    raise ValueError(
                        f"Stream item did not start with `{self.SSE_DATA_PREFIX}`. Was `{stream_item_as_str}`"
                    )

                payload = stream_item_as_str[len(self.SSE_DATA_PREFIX) :]
                if payload == "[DONE]":
                    continue

                yield json.loads(payload)

    def _build_query_parameters(self) -> Mapping[str, str]:
        return {
            # cannot use str() here because we want lowercase true/false in query string
            # Also do not want to send the nice flag with every request if it is false
            **({"nice": "true"} if self.nice else {}),
        }

    def _build_json_body(
        self, request: AnyRequest, model: Optional[str]
    ) -> Dict[str, Any]:
        json_body = dict(request.to_json())

        if model is not None:
            json_body["model"] = model
        if self.hosting is not None:
            json_body["hosting"] = self.hosting
        if self.tags is not None:
            json_body["tags"] = self.tags
        return json_body

    async def models(self) -> List[Mapping[str, Any]]:
        """
        Queries all models which are currently available.

        For documentation of the response, see
        https://docs.aleph-alpha.com/products/apis/pharia-inference/available-models/
        """
        return await self._get_request_json("models_available")  # type: ignore

    async def complete(
        self,
        request: CompletionRequest,
        model: str,
    ) -> CompletionResponse:
        """Generates completions given a prompt.

        Parameters:
            request (CompletionRequest, required):
                Parameters for the requested completion.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # create a prompt
            >>> prompt = Prompt.from_text("An apple a day, ")
            >>>
            >>> # create a completion request
            >>> request = CompletionRequest(
                    prompt=prompt,
                    maximum_tokens=32,
                    stop_sequences=["###","\\n"],
                    temperature=0.12
                )
            >>>
            >>> # complete the prompt
            >>> result = await client.complete(request, model=model_name)
        """
        response = await self._post_request(
            "complete",
            request,
            model,
        )
        return CompletionResponse.from_json(response)

    async def chat(
        self,
        request: ChatRequest,
        model: str,
    ) -> ChatResponse:
        """Chat with a model.

        Parameters:
            request (ChatRequest, required):
                Parameters for the requested chat.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> import os
            >>> from aleph_alpha_client import AsyncClient, ChatRequest, Message
            >>>
            >>> client = AsyncClient(token=os.environ["TOKEN"], host="https://inference-api.your.domain")
            >>> model = "llama-3.1-8b-instruct"
            >>> # create a chat request
            >>> request = ChatRequest(
                    messages=[Message(role="user", content="Hello, how are you?")],
                    model=model,
                )
            >>>
            >>> # chat with the model
            >>> result = await client.chat(request, model=model)
            >>> print(result.message)
        """
        response = await self._post_request(
            "chat/completions",
            request,
            model,
        )
        return ChatResponse.from_json(response)

    async def complete_with_streaming(
        self,
        request: CompletionRequest,
        model: str,
    ) -> AsyncGenerator[CompletionResponseStreamItem, None]:
        """Generates streamed completions given a prompt.

        Parameters:
            request (CompletionRequest, required):
                Parameters for the requested completion.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # create a prompt
            >>> prompt = Prompt.from_text("An apple a day, ")
            >>>
            >>> # create a completion request
            >>> request = CompletionRequest(
                    prompt=prompt,
                    maximum_tokens=32,
                    stop_sequences=["###","\\n"],
                    temperature=0.12
                )
            >>>
            >>> # complete the prompt
            >>> result = await client.complete_with_streaming(request, model=model_name)
            >>>
            >>> # consume the completion stream
            >>> async for stream_item in result:
            >>>     do_something_with(stream_item)
        """
        async for stream_item_json in self._post_request_with_streaming(
            "complete",
            request,
            model,
        ):
            yield stream_item_from_json(stream_item_json)

    async def chat_with_streaming(
        self,
        request: ChatRequest,
        model: str,
    ) -> AsyncGenerator[Union[ChatStreamChunk, Usage, ToolCall, FinishReason], None]:
        """Generates streamed chat completions.

        The first yielded chunk contains the role, while subsequent chunks only contain the content delta.

        Parameters:
            request (ChatRequest, required):
                Parameters for the requested chat.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> import os
            >>> from aleph_alpha_client import AsyncClient, ChatRequest, Message
            >>>
            >>> client = AsyncClient(token=os.environ["TOKEN"], host="https://inference-api.your.domain")
            >>> model = "llama-3.1-8b-instruct"
            >>> # create a chat request
            >>> request = ChatRequest(
                    messages=[Message(role="user", content="Hello, how are you?")],
                    model=model,
                )
            >>>
            >>> # chat with the model
            >>> result = client.chat_with_streaming(request, model=model)
            >>>
            >>> # consume the chat stream
            >>> async for stream_item in result:
            >>>     print(stream_item)
        """

        async for x in process_chat_stream(
            self._post_request_with_streaming(
                "chat/completions",
                request,
                model,
            )
        ):
            yield x

    async def tokenize(
        self,
        request: TokenizationRequest,
        model: str,
    ) -> TokenizationResponse:
        """Tokenizes the given prompt for the given model.

        Parameters:
            request (TokenizationRequest, required):
                Parameters for the requested tokenization.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = TokenizationRequest(prompt="hello", token_ids=True, tokens=True)
            >>> response = await client.tokenize(request, model=model_name)
        """
        response = await self._post_request(
            "tokenize",
            request,
            model,
        )
        return TokenizationResponse.from_json(response)

    async def detokenize(
        self,
        request: DetokenizationRequest,
        model: str,
    ) -> DetokenizationResponse:
        """Detokenizes the given prompt for the given model.

        Parameters:
            request (DetokenizationRequest, required):
                Parameters for the requested detokenization.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = DetokenizationRequest(token_ids=[2, 3, 4])
            >>> response = await client.detokenize(request, model=model_name)
        """
        response = await self._post_request(
            "detokenize",
            request,
            model,
        )
        return DetokenizationResponse.from_json(response)

    async def embed(
        self,
        request: EmbeddingRequest,
        model: str,
    ) -> EmbeddingResponse:
        """Embeds a text and returns vectors that can be used for downstream tasks (e.g. semantic similarity) and models (e.g. classifiers).

        Parameters:
            request (EmbeddingRequest, required):
                Parameters for the requested embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = EmbeddingRequest(prompt=Prompt.from_text("This is an example."), layers=[-1], pooling=["mean"])
            >>> result = await client.embed(request, model=model_name)
        """
        response = await self._post_request(
            "embed",
            request,
            model,
        )
        return EmbeddingResponse.from_json(response)

    async def semantic_embed(
        self,
        request: SemanticEmbeddingRequest,
        model: str,
    ) -> SemanticEmbeddingResponse:
        """Embeds a text and returns vectors that can be used for downstream tasks
        (e.g. semantic similarity) and models (e.g. classifiers).

        Parameters:
            request (SemanticEmbeddingRequest, required):
                Parameters for the requested semantic embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # function for symmetric embedding
            >>> async def embed_symmetric(text: str):
                    # Create an embeddingrequest with the type set to symmetric
                    request = SemanticEmbeddingRequest(prompt=Prompt.from_text(text), representation=SemanticRepresentation.Symmetric)
                    # create the embedding
                    result = await client.semantic_embed(request, model=model_name)
                    return result.embedding
            >>>
            >>> # function to calculate similarity
            >>> def cosine_similarity(v1: Sequence[float], v2: Sequence[float]) -> float:
                    "compute cosine similarity of v1 to v2: (v1 dot v2)/{||v1||*||v2||)"
                    sumxx, sumxy, sumyy = 0, 0, 0
                    for i in range(len(v1)):
                        x = v1[i]; y = v2[i]
                        sumxx += x*x
                        sumyy += y*y
                        sumxy += x*y
                    return sumxy/math.sqrt(sumxx*sumyy)
            >>>
            >>> # define the texts
            >>> text_a = "The sun is shining"
            >>> text_b = "Il sole splende"
            >>>
            >>> # show the similarity
            >>> print(cosine_similarity(await embed_symmetric(text_a), await embed_symmetric(text_b)))
        """
        response = await self._post_request(
            "semantic_embed",
            request,
            model,
        )
        return SemanticEmbeddingResponse.from_json(response)

    async def batch_semantic_embed(
        self,
        request: BatchSemanticEmbeddingRequest,
        model: Optional[str] = None,
        num_concurrent_requests: int = 1,
        batch_size: int = 100,
        progress_bar: bool = False,
    ) -> BatchSemanticEmbeddingResponse:
        """Embeds a sequence of texts or images and returns vectors in the same order as they
        were provided. If more than `batch_size` prompts are provided then this method will chunk them
        into batches of up to `batch_size` prompts that will be sent to the API.

        Parameters:
            request (BatchSemanticEmbeddingRequest, required):
                Parameters for the requested semantic embeddings.

            model (string, optional, default None):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

            num_concurrent_requests (int, optional, default 1):
                Maximum number of concurrent requests to send to the API.

            batch_size (int, optional, default 100):
                Number of prompts per batch sent to the API. This value must be between 1 and 100 (inclusive).

            progress_bar (bool, optional, default False):
                Whether to show a progress bar using tqdm.

        Examples:
            >>> # function for symmetric embedding
            >>> def embed_symmetric(texts: Sequence[str]):
                    # Create an embeddingrequest with the type set to symmetric
                    request = BatchSemanticEmbeddingRequest(
                        prompts=[Prompt.from_text(text) for text in texts],
                        representation=SemanticRepresentation.Symmetric
                    )
                    # create the embedding
                    result = client.batch_semantic_embed(request, model=model_name)
                    return result.embedding
        """
        if batch_size < 1 or batch_size > 100:
            raise ValueError(
                "`batch_semantic_embed` must be called with a `batch_size` between 1 and 100 (inclusive)"
            )

        responses: List[EmbeddingVector] = []
        model_version = ""

        # The API currently only supports batch semantic embedding requests with up to 100
        # prompts per batch. As a convenience for users, this function chunks larger requests.
        results = await self._gather_with_concurrency(
            "batch_semantic_embed",
            model,
            num_concurrent_requests,
            _generate_semantic_embedding_batches(request, batch_size),
            progress_bar,
        )
        num_tokens_prompt_total = 0
        for result in results:
            resp = BatchSemanticEmbeddingResponse.from_json(result)
            model_version = resp.model_version
            responses.extend(resp.embeddings)
            num_tokens_prompt_total += resp.num_tokens_prompt_total

        return BatchSemanticEmbeddingResponse(
            model_version=model_version,
            embeddings=responses,
            num_tokens_prompt_total=num_tokens_prompt_total,
        )

    async def instructable_embed(
        self,
        request: InstructableEmbeddingRequest,
        model: str,
    ) -> InstructableEmbeddingResponse:
        """Embeds a text and returns vectors that can be used for classification according to a given instruction.

        Parameters:
            request (InstructableEmbeddingRequest, required):
                Parameters for the requested instructable embedding.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> # function for salutation embedding
            >>> async def embed_salutation(text: str):
                    # Create an embeddingrequest with a given instruction
                    request = InstructableEmbeddingRequest(
                        input=Prompt.from_text(text),
                        instruction="Represent the text to query a database of salutations"
                    )
                    # create the embedding
                    result = await client.instructable_embed(request, model=model_name)
                    return result.embedding
            >>>
            >>> # function to calculate similarity
            >>> def cosine_similarity(v1: Sequence[float], v2: Sequence[float]) -> float:
                    "compute cosine similarity of v1 to v2: (v1 dot v2)/{||v1||*||v2||)"
                    sumxx, sumxy, sumyy = 0, 0, 0
                    for i in range(len(v1)):
                        x = v1[i]; y = v2[i]
                        sumxx += x*x
                        sumyy += y*y
                        sumxy += x*y
                    return sumxy/math.sqrt(sumxx*sumyy)
            >>>
            >>> # define the texts
            >>> text_a = "Hello"
            >>> text_b = "Good morning"
            >>>
            >>> # show the similarity
            >>> print(cosine_similarity(await embed_salutation(text_a), await embed_salutation(text_b)))
        """
        response = await self._post_request(
            "instructable_embed",
            request,
            model,
        )
        return InstructableEmbeddingResponse.from_json(response)

    async def evaluate(
        self,
        request: EvaluationRequest,
        model: str,
    ) -> EvaluationResponse:
        """Evaluates the model's likelihood to produce a completion given a prompt.

        Parameters:
            request (EvaluationRequest, required):
                Parameters for the requested evaluation.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = EvaluationRequest(
                    prompt=Prompt.from_text("hello"), completion_expected=" world"
                )
            >>> response = await client.evaluate(request, model=model_name)
        """
        response = await self._post_request(
            "evaluate",
            request,
            model,
        )
        return EvaluationResponse.from_json(response)

    async def explain(
        self,
        request: ExplanationRequest,
        model: str,
    ) -> ExplanationResponse:
        """Better understand the source of a completion, specifically on how much each section of a
        prompt impacts each token of the completion.

        Parameters:
            request (ExplanationRequest, required):
                Parameters for the requested explanation.

            model (string, required):
                Name of model to use. A model name refers to a model architecture (number of parameters among others).
                Always the latest version of model is used.

        Examples:
            >>> request = ExplanationRequest(
                    prompt=Prompt.from_text("Andreas likes"),
                    target=" pizza."
                )
            >>> response = await client.explain(request, model="luminous-base")
        """
        response = await self._post_request(
            "explain",
            request,
            model,
        )
        return ExplanationResponse.from_json(response)

    async def create_steering_concept(
        self, request: SteeringConceptCreationRequest
    ) -> SteeringConceptCreationResponse:
        """Creates a steering concept.

        A steering concept consists of a list of steering examples. A steering
        example is a pair of a "negative" and a "positive" string, describing
        how you want to alter the model's output.

        This request will return a unique ID for the newly created steering
        concept that you can then use in completion and chat requests.

        Parameters:
            request (SteeringConceptCreationRequest, required)
                Parameters for the steering concepts to create.

        Examples:
            >>> request = SteeringConceptCreationRequest(
            >>>     examples=[
            >>>         SteeringPairedExample(
            >>>             negative="I appreciate your valuable feedback on this matter.",
            >>>             positive="Thanks for the real talk, fam.",
            >>>         ),
            >>>         SteeringPairedExample(
            >>>             negative="The financial projections indicate significant growth potential.",
            >>>             positive="Yo, these numbers are looking mad stacked!",
            >>>         ),
            >>>     ]
            >>> )
            >>> response = client.create_steering_concept(request)
        """
        response = await self._post_request(
            "steering_concepts",
            request,
        )
        return SteeringConceptCreationResponse.from_json(response)

    async def tokenizer(self, model: str) -> Tokenizer:
        """Returns a Tokenizer instance with the settings that were used to train the model.

        Examples:
            >>> tokenizer = await client.tokenizer(model="luminous-base")
            >>> tokenized_prompt = tokenizer.encode("Hello world")
        """
        response = await self._get_request_text(f"models/{model}/tokenizer")
        return Tokenizer.from_str(response)

    # Based on: https://docs.aleph-alpha.com/changelog/2022/11/14/async-python-client/
    async def _gather_with_concurrency(
        self,
        endpoint: str,
        model: Optional[str],
        n: int,
        requests: Sequence[BatchSemanticEmbeddingRequest],
        progress_bar: bool,
    ) -> List[Dict[str, Any]]:
        semaphore = asyncio.Semaphore(n)

        async def sem_task(request: BatchSemanticEmbeddingRequest):
            async with semaphore:
                return await self._post_request(endpoint, request, model)

        # asyncio.gather preserves order of awaitables in result list
        if progress_bar:
            return await tqdm.gather(*(sem_task(request) for request in requests))
        else:
            return await asyncio.gather(*(sem_task(request) for request in requests))

    async def translate(
        self,
        request: TranslationRequest,
    ) -> TranslationResponse:
        """Translates text from one language to another.

        Parameters:
            request (TranslationRequest, required):
                Parameters for the requested translation.

        Examples:
            >>> request = TranslationRequest(
                    model="pharia-1-mt-translation",
                    source="Hello, how are you?",
                    target_language="de"
                )
            >>> response = await client.translate(request)
            >>> print(response.translation)
        """
        response = await self._post_request("translate", request)
        return TranslationResponse.from_json(response)

    async def rerank(
        self,
        request: RerankRequest,
        model: str,
    ) -> RerankResponse:
        """Reranks documents against a query.

        This endpoint takes in a query and a list of documents and produces an array
        with each document assigned a relevance score.

        Parameters:
            request (RerankRequest, required):
                Parameters for the requested reranking.

            model (string, required):
                Name of the model to use for reranking.

        Examples:
            >>> request = RerankRequest(
                    query="What is the capital of France?",
                    documents=[
                        "The capital of Brazil is Brasilia.",
                        "The capital of France is Paris.",
                        "Horses and cows are both animals.",
                    ],
                    top_n=2,
                )
            >>> response = await client.rerank(request, model="your-reranker-model")
            >>> for result in response.results:
            >>>     print(f"Document {result.index}: {result.relevance_score}")
        """
        response = await self._post_request("rerank", request, model)
        return RerankResponse.from_json(response)


def _generate_semantic_embedding_batches(
    request: BatchSemanticEmbeddingRequest, batch_size: int = 100
) -> List[BatchSemanticEmbeddingRequest]:
    requests = []
    for batch_index in range(0, len(request.prompts), batch_size):
        batch = request.prompts[batch_index : batch_index + batch_size]
        requests.append(
            BatchSemanticEmbeddingRequest(
                prompts=batch,
                representation=request.representation,
                compress_to_size=request.compress_to_size,
                normalize=request.normalize,
                contextual_control_threshold=request.contextual_control_threshold,
                control_log_additive=request.control_log_additive,
            )
        )
    return requests



import json
from pathlib import Path
from typing import List, Union

import pytest
from PIL import Image

from aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.chat import (
    ChatRequest,
    ChatStreamChunk,
    FinishReason,
    Message,
    Role,
    StreamOptions,
    TextMessage,
    Usage,
    stream_chat_item_from_json,
    ToolCall,
    FunctionCall,
)
from aleph_alpha_client.structured_output import JSONSchema
from pydantic import BaseModel

from .conftest import GenericClient
from .test_steering import create_sample_steering_concept_creation_request


@pytest.mark.vcr
async def test_can_not_chat_with_all_models(async_client: AsyncClient, model_name: str):
    request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=model_name,
        maximum_tokens=7,
    )

    with pytest.raises(ValueError):
        await async_client.chat(request, model=model_name)


@pytest.mark.vcr
def test_can_chat_with_chat_model(sync_client: Client, chat_model_name: str):
    request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=chat_model_name,
    )

    response = sync_client.chat(request, model=chat_model_name)
    assert response.message.role == Role.Assistant
    assert response.message.content is not None
    assert isinstance(response.finish_reason, FinishReason)


@pytest.mark.vcr
async def test_can_chat_with_async_client(
    async_client: AsyncClient, chat_model_name: str
):
    system_msg = Message(role=Role.System, content="You are a helpful assistant.")
    user_msg = Message(role=Role.User, content="Hello, how are you?")
    request = ChatRequest(
        messages=[system_msg, user_msg],
        model=chat_model_name,
    )

    response = await async_client.chat(request, model=chat_model_name)
    assert response.message.role == Role.Assistant
    assert response.message.content is not None


TOOLS = [
    {
        "type": "function",
        "function": {
            "name": "get_weather",
            "description": "Get current temperature for a given location.",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "City and country e.g. Bogotá, Colombia",
                    }
                },
                "required": ["location"],
                "additionalProperties": False,
            },
            "strict": True,
        },
    }
]


@pytest.mark.vcr
async def test_can_chat_with_tools(
    async_client: AsyncClient, tool_calling_model_name: str
):
    system_msg = Message(role=Role.System, content="You are a helpful assistant.")
    user_msg = Message(
        role=Role.User, content="What is the weather like in Paris today?"
    )
    request = ChatRequest(
        messages=[system_msg, user_msg],
        model=tool_calling_model_name,
        tools=TOOLS,
    )

    response = await async_client.chat(request, model=tool_calling_model_name)
    assert response.message.role == Role.Assistant
    assert response.message.content is not None
    assert response.message.tool_calls is not None
    calls = response.message.tool_calls
    assert len(calls) == 1
    assert calls[0].type == "function"
    assert calls[0].function.name == "get_weather"

    request = ChatRequest(
        messages=[
            system_msg,
            Message(
                role=response.message.role,
                content=response.message.content,
                tool_calls=response.message.tool_calls,
            ),
            Message(
                role=Role.Tool,
                content="Cloudy with a bit of rain.",
                tool_call_id=response.message.tool_calls[0].id,
            ),
            user_msg,
        ],
        model=tool_calling_model_name,
        tools=TOOLS,
    )

    response = await async_client.chat(request, model=tool_calling_model_name)
    assert "cloudy" in response.message.content.lower()


@pytest.mark.vcr
async def test_can_chat_with_streaming_support(
    async_client: AsyncClient, chat_model_name: str
):
    request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=chat_model_name,
    )

    stream = async_client.chat_with_streaming(request, model=chat_model_name)
    stream_items = [stream_item async for stream_item in stream]

    first = stream_items[0]
    assert isinstance(first, ChatStreamChunk) and first.role is not None
    assert all(
        isinstance(item, ChatStreamChunk) and item.content is not None
        for item in stream_items[1:-1]
    )
    assert isinstance(stream_items[-1], FinishReason)


@pytest.mark.vcr
async def test_can_chat_with_tools_streamed(
    async_client: AsyncClient, tool_calling_model_name: str
):
    system_msg = Message(role=Role.System, content="You are a helpful assistant.")
    user_msg = Message(
        role=Role.User, content="What is the weather like in Paris today?"
    )
    request = ChatRequest(
        messages=[system_msg, user_msg],
        model=tool_calling_model_name,
        tools=TOOLS,
    )

    stream = async_client.chat_with_streaming(request, model=tool_calling_model_name)
    stream_items = [stream_item async for stream_item in stream]

    v = stream_items[-2]
    tool_call_id = v.id if isinstance(v, ToolCall) else ""

    assert stream_items[-3:] == [
        ChatStreamChunk(
            content="</think>\n\n",
            role=None,
            tool_calls=None,
        ),
        ToolCall(
            id=tool_call_id,
            type="function",
            function=FunctionCall(
                name="get_weather",
                arguments='{"location": "Paris, France"}',
            ),
        ),
        FinishReason.ToolCalls,
    ]


def test_usage_response_is_parsed():
    # Given an API response with usage data and no choice
    data = {
        "choices": [],
        "created": 1730133402,
        "model": "llama-3.1-70b-instruct",
        "system_fingerprint": ".unknown.",
        "object": "chat.completion.chunk",
        "usage": {"prompt_tokens": 31, "completion_tokens": 88, "total_tokens": 119},
    }

    # When parsing it
    result = stream_chat_item_from_json(data)

    # Then a usage instance is returned
    assert isinstance(result, Usage)
    assert result.prompt_tokens == 31


def test_finish_reason_response_is_parsed():
    # Given an API response with finish reason and no choice
    data = {
        "choices": [
            {
                "finish_reason": "stop",
                "index": 0,
                "delta": {},
                "logprobs": None,
            }
        ],
        "created": 1730133402,
        "model": "llama-3.1-70b-instruct",
        "system_fingerprint": ".unknown.",
        "object": "chat.completion.chunk",
    }

    # When parsing it
    result = stream_chat_item_from_json(data)

    # Then a finish reason instance is returned
    assert isinstance(result, FinishReason)
    assert result == FinishReason.Stop


def test_chunk_response_is_parsed():
    # Given an API response without usage data
    data = {
        "choices": [
            {
                "finish_reason": None,
                "index": 0,
                "delta": {"content": " way, those clothes you're wearing"},
                "logprobs": None,
            }
        ],
        "created": 1730133401,
        "model": "llama-3.1-70b-instruct",
        "system_fingerprint": None,
        "object": "chat.completion.chunk",
        "usage": None,
    }

    # When parsing it
    result = stream_chat_item_from_json(data)

    # Then a ChatStreamChunk instance is returned
    assert isinstance(result, ChatStreamChunk)
    assert result.content == " way, those clothes you're wearing"


@pytest.mark.vcr
async def test_stream_options(async_client: AsyncClient, chat_model_name: str):
    # Given a request with include usage options set
    stream_options = StreamOptions(include_usage=True)
    request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=chat_model_name,
        stream_options=stream_options,
    )

    # When receiving the chunks
    stream = async_client.chat_with_streaming(request, model=chat_model_name)
    stream_items = [stream_item async for stream_item in stream]

    print(stream_items)
    # Then
    assert all(isinstance(item, ChatStreamChunk) for item in stream_items[:-2])
    assert isinstance(stream_items[-2], FinishReason)
    assert isinstance(stream_items[-1], Usage)


@pytest.mark.vcr
def test_steering_chat(sync_client: Client, chat_model_name: str):
    base_request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=chat_model_name,
    )

    steering_concept_id = sync_client.create_steering_concept(
        create_sample_steering_concept_creation_request()
    ).id

    steered_request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model=chat_model_name,
        steering_concepts=[steering_concept_id],
    )

    base_response = sync_client.chat(base_request, model=chat_model_name)
    steered_response = sync_client.chat(steered_request, model=chat_model_name)

    base_completion_result = base_response.message.content
    steered_completion_result = steered_response.message.content

    assert base_completion_result
    assert steered_completion_result
    assert base_completion_result != steered_completion_result


def remove_thinking_content(content: str) -> str:
    needle = "</think>"
    thinking = content.find(needle)
    content = content if thinking == -1 else content[thinking + len(needle) :]
    return content


@pytest.mark.vcr
def test_response_format_json_schema(
    sync_client: Client, structured_output_model_name: str
):
    example_json_schema = {
        "type": "object",
        "title": "Aquarium",
        "properties": {
            "nemo": {
                "type": "string",
                "title": "Nemo",
                "description": "Name of the fish",
            },
            "species": {
                "type": "string",
                "title": "Species",
                "description": "The species of the fish (e.g., Clownfish, Goldfish)",
            },
            "color": {
                "type": "string",
                "title": "Color",
                "description": "Primary color of the fish",
            },
            "size_cm": {
                "type": "number",
                "title": "Size in centimeters",
                "description": "Length of the fish in centimeters",
                "minimum": 0.1,
                "maximum": 100.0,
            },
        },
        "required": ["nemo", "species", "color", "size_cm"],
    }

    request = ChatRequest(
        messages=[
            Message(role=Role.System, content="You are a helpful assistant."),
            Message(
                role=Role.User,
                content=f"Give me JSON {example_json_schema}! Tell me about nemo",
            ),
        ],
        model=structured_output_model_name,
        response_format=JSONSchema(
            schema=example_json_schema,
            name="aquarium",
            description="Describe nemo",
            strict=True,
        ),
    )

    response = sync_client.chat(request, model=structured_output_model_name)
    json_response = json.loads(remove_thinking_content(response.message.content))
    # Validate all required fields are present
    required_fields = ["nemo", "species", "color", "size_cm"]
    for field in required_fields:
        assert (
            field in json_response.keys()
        ), f"Required field '{field}' is missing from response"
    # Validate field types
    assert isinstance(json_response["nemo"], str), "Field 'nemo' should be a string"
    assert isinstance(
        json_response["species"], str
    ), "Field 'species' should be a string"
    assert isinstance(json_response["color"], str), "Field 'color' should be a string"
    assert isinstance(
        json_response["size_cm"], (int, float)
    ), "Field 'size_cm' should be a number"
    # Validate size constraints
    assert (
        0.1 <= json_response["size_cm"] <= 100.0
    ), "Field 'size_cm' should be between 0.1 and 100.0"


@pytest.mark.vcr
def test_response_format_json_schema_pydantic(
    sync_client: Client, structured_output_model_name: str
):
    class Aquarium(BaseModel):
        nemo: str
        species: str
        color: str
        size_cm: float

    request = ChatRequest(
        messages=[
            Message(role=Role.System, content="You are a helpful assistant."),
            Message(
                role=Role.User,
                content=f"Give me JSON of type {Aquarium}! Tell me about nemo",
            ),
        ],
        model=structured_output_model_name,
        response_format=Aquarium,
    )

    response = sync_client.chat(request, model=structured_output_model_name)
    content = remove_thinking_content(response.message.content)
    # Tests that it is valid json and loads
    json.loads(content)

    # Validate against desired fields
    Aquarium.model_validate_json(content)


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
async def test_can_chat_with_images(
    generic_client: GenericClient, dummy_model_name: str
):
    image_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    image = Image.open(image_path)

    request = ChatRequest(
        messages=[
            Message(
                role=Role.User,
                content=[
                    "Describe the following image.",
                    image,
                ],
            )
        ],
        model=dummy_model_name,
        maximum_tokens=200,
    )
    response = await generic_client.chat(request, model=dummy_model_name)

    # If the dummy worker receives images, it returns their dimensions in pixels
    # as token ids. Currently, the scheduler will crop and resize the images to
    # 384x384 pixels.
    #
    # el = 384
    assert response.message.content == "el el"


TINY_PNG = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR4nGNgYGD4DwABBAEAX+XDSwAAAABJRU5ErkJggg=="


def test_multimodal_message_serialization() -> None:
    image_path = Path(__file__).parent / "tiny.png"
    message = Message(
        role=Role.User,
        content=[
            "Describe the following image.",
            Image.open(image_path),
        ],
    )
    assert message.to_json() == {
        "role": "user",
        "content": [
            {"type": "text", "text": "Describe the following image."},
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/png;base64,{TINY_PNG}"},
            },
        ],
    }


def test_multimodal_message_serialization_unknown_type() -> None:
    image_path = Path(__file__).parent / "tiny.png"
    message = Message(
        role=Role.User,
        content=[
            "Describe the following image.",
            Path(image_path),  # type: ignore
        ],
    )
    with pytest.raises(ValueError) as e:
        message.to_json()
    assert (
        str(e.value)
        == "The item in the prompt is not valid. Try either a string or an Image."
    )


def test_request_serialization_no_default_values() -> None:
    request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello, how are you?")],
        model="dummy-model",
    )
    assert request.to_json() == {
        "model": "dummy-model",
        "messages": [{"role": "user", "content": "Hello, how are you?"}],
    }


def test_multi_turn_chat_serialization(sync_client: Client, chat_model_name: str):
    """
    Test that TextMessage can be serialized when included in multi-turn chat history.

    We previously encountered an error in a multi-turn chat conversation.
    The returned TextMessage could not be made part of the chat history for the
    next request as the method for serialization was missing.
    This test should catch such conversion issues.
    """

    # First turn
    first_request = ChatRequest(
        messages=[Message(role=Role.User, content="Hello")],
        model=chat_model_name,
    )
    first_response = sync_client.chat(first_request, model=chat_model_name)
    # Second turn - includes the TextMessage from first response in history
    messages_with_history: List[Union[Message, TextMessage]] = [
        Message(role=Role.User, content="Hello"),
        first_response.message,  # This TextMessage must be serializable
        Message(role=Role.User, content="Follow up question"),
    ]

    second_request = ChatRequest(
        messages=messages_with_history,
        model=chat_model_name,
    )

    # This would fail if TextMessage.to_json() doesn't exist
    serialized = second_request.to_json()
    assert len(serialized["messages"]) == 3
    assert serialized["messages"][1]["role"] == "assistant"



import asyncio
import pytest
from aleph_alpha_client import AsyncClient
from aleph_alpha_client.translation import TranslationRequest, TranslationResponse
from tests.conftest import GenericClient


LONG_ENGLISH_TEXT = 'Question: What are some benefits of surface micro-machining?\nAnswer the question on the basis of the text. If there is no answer within the text, respond "no answer in text".\n\nSurface micromachining\n\nSurface micromachining builds microstructures by deposition and etching structural layers over a substrate.[1] This is different from Bulk micromachining, in which a silicon substrate wafer is selectively etched to produce structures.\n\nLayers\n\nGenerally, polysilicon is used as one of the substrate layers while silicon dioxide is used as a sacrificial layer. The sacrificial layer is removed or etched out to create any necessary void in the thickness direction. Added layers tend to vary in size from 2-5 micrometres. The main advantage of this machining process is the ability to build electronic and mechanical components (functions) on the same substrate. Surface micro-machined components are smaller compared to their bulk micro-machined counterparts.\n\nAs the structures are built on top of the substrate and not inside it, the substrate\'s properties are not as important as in bulk micro-machining. Expensive silicon wafers can be replaced by cheaper substrates, such as glass or plastic. The size of the substrates may be larger than a silicon wafer, and surface micro-machining is used to produce thin-film transistors on large area glass substrates for flat panel displays. This technology can also be used for the manufacture of thin film solar cells, which can be deposited on glass, polyethylene terepthalate substrates or other non-rigid materials.\n\nFabrication process\n\nMicro-machining starts with a silicon wafer or other substrate upon which new layers are grown. These layers are selectively etched by photo-lithography; either a wet etch involving an acid, or a dry etch involving an ionized gas (or plasma). Dry etching can combine chemical etching with physical etching or ion bombardment. Surface micro-machining involves as many layers as are needed with a different mask (producing a different pattern) on each layer. Modern integrated circuit fabrication uses this technique and can use as many as 100 layers. Micro-machining is a younger technology and usually uses no more than 5 or 6 layers. Surface micro-machining uses developed technology (although sometimes not enough for demanding applications) which is easily repeatable for volume production.\n\n### Response:'


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
async def test_can_translate_single_request(
    generic_client: GenericClient, translation_model_name: str
):
    request = TranslationRequest(
        model=translation_model_name,
        source=LONG_ENGLISH_TEXT,
        target_language="de",
    )

    response = await generic_client.translate(request)
    assert isinstance(response, TranslationResponse)
    assert response.translation is not None
    assert response.score >= 0.0
    assert response.score <= 1.0
    assert response.num_tokens_prompt_total > 0
    assert response.num_tokens_generated == 0  # Translation models don't expose tokens

    if response.segments is not None:
        for segment in response.segments:
            assert segment.source is not None
            assert segment.translation is not None
            assert segment.score >= 0.0
            assert segment.score <= 1.0


@pytest.mark.vcr
async def test_can_translate_multiple_requests_concurrently_with_async_client(
    async_client: AsyncClient, translation_model_name: str
):
    # Create multiple translation requests for different languages
    languages = ["de", "fr", "es", "it", "nl"]
    requests = [
        TranslationRequest(
            model=translation_model_name,
            source=LONG_ENGLISH_TEXT,
            target_language=lang,
        )
        for lang in languages
    ]

    responses = await asyncio.gather(
        *[async_client.translate(request) for request in requests]
    )

    assert len(responses) == len(languages)
    for response in responses:
        assert isinstance(response, TranslationResponse)
        assert response.translation is not None
        assert response.translation != LONG_ENGLISH_TEXT
        assert response.score >= 0.0
        assert response.score <= 1.0
        assert response.num_tokens_prompt_total > 0
        assert response.num_tokens_generated == 0

        if response.segments is not None:
            for segment in response.segments:
                assert segment.source is not None
                assert segment.translation is not None
                assert segment.score >= 0.0
                assert segment.score <= 1.0



from dataclasses import dataclass, asdict
from typing import Any, Dict, Mapping, Sequence


@dataclass(frozen=True)
class DetokenizationRequest:
    """Describes a detokenization request.

    Parameters
        token_ids (Sequence[int])
            Ids of the tokens for which the text should be returned.

    Examples:
        >>> DetokenizationRequest(token_ids=[1730, 387, 300, 4377, 17])
    """

    token_ids: Sequence[int]

    def to_json(self) -> Mapping[str, Any]:
        payload = self._asdict()
        return payload

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class DetokenizationResponse:
    result: str

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "DetokenizationResponse":
        return DetokenizationResponse(result=json["result"])



from aleph_alpha_client.aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.detokenization import DetokenizationRequest
import pytest

# AsyncClient


@pytest.mark.vcr
async def test_can_detokenization_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = DetokenizationRequest(token_ids=[2, 3, 4])

    response = await async_client.detokenize(request, model=model_name)
    assert len(response.result) > 0


# Client


@pytest.mark.vcr
def test_detokenize(sync_client: Client, model_name: str):
    response = sync_client.detokenize(DetokenizationRequest([4711]), model=model_name)

    assert response.result is not None



from http import HTTPStatus
from pathlib import Path

from pytest import raises
from pytest_httpserver import HTTPServer
from requests import RequestException

from aleph_alpha_client import Image


def test_from_url_with_non_OK_response(httpserver: HTTPServer):
    path = "/image"
    httpserver.expect_request(path).respond_with_data(
        "html", status=HTTPStatus.FORBIDDEN
    )

    with raises(RequestException):
        Image.from_url(httpserver.url_for(path))


def test_from_image_source_on_local_file_path_pathlib():
    image_source_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    str_image_source_path = str(image_source_path)

    img_1 = Image.from_image_source(image_source=image_source_path)
    img_2 = Image.from_file(path=str_image_source_path)
    assert img_1.base_64 == img_2.base_64


def test_from_image_source_on_local_file_path_str():
    image_source_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    image_source_path = str(image_source_path)
    img_1 = Image.from_image_source(image_source=image_source_path)
    img_2 = Image.from_file(path=image_source_path)
    assert img_1.base_64 == img_2.base_64


def test_from_image_source_on_url():
    image_source_url = "https://docs.aleph-alpha.com/COMMON/_images/pharia-inference-complete-endpoint-1.jpg"
    img_1 = Image.from_image_source(image_source=image_source_url)
    img_2 = Image.from_url(url=image_source_url)
    assert img_1.base_64 == img_2.base_64


def test_from_image_source_on_bytes():
    image_source_bytes = (Path(__file__).parent / "dog-and-cat-cover.jpg").read_bytes()
    img_1 = Image.from_image_source(image_source=image_source_bytes)
    img_2 = Image.from_bytes(bytes=image_source_bytes)
    assert img_1.base_64 == img_2.base_64


def test_from_image_source_on_input_of_a_wrong_type():
    with raises(TypeError):
        Image.from_image_source(image_source=None)



import base64
import dataclasses
from dataclasses import asdict, dataclass
from enum import Enum
from io import BytesIO
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    Literal,
    AsyncGenerator,
)

from pydantic import BaseModel
from aleph_alpha_client.structured_output import ResponseFormat
from PIL.Image import Image


class Role(str, Enum):
    """A role used for a message in a chat."""

    User = "user"
    Assistant = "assistant"
    System = "system"
    Tool = "tool"


@dataclass(frozen=True)
class FunctionCall:
    name: str
    arguments: str


@dataclass(frozen=True)
class ToolCall:
    id: str
    type: str
    function: FunctionCall

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "ToolCall":
        function = json["function"]
        return ToolCall(
            id=json["id"],
            type=json["type"],
            function=FunctionCall(
                name=function["name"], arguments=function["arguments"]
            ),
        )

    def to_json(self) -> Mapping[str, Any]:
        return {
            "id": self.id,
            "type": self.type,
            "function": {
                "name": self.function.name,
                "arguments": self.function.arguments,
            },
        }


@dataclass(frozen=True)
class Message:
    """
    Describes a message in a chat.

    Parameters:
        role (Role, required):
            The role of the message.

        content (str | List[Union[str | Image]], required):
            The content of the message.
    """

    role: Role
    content: Union[str, List[Union[str, Image]]]
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None

    def to_json(self) -> Mapping[str, Any]:
        result = {
            "role": self.role.value,
            "content": _message_content_to_json(self.content),
        }
        if self.tool_calls is not None:
            result["tool_calls"] = [t.to_json() for t in self.tool_calls]
        if self.tool_call_id is not None:
            result["tool_call_id"] = self.tool_call_id
        return result


# We introduce a more specific message type because chat responses can only
# contain text at the moment. This enables static type checking to proof that
# `content` is always a string.
@dataclass(frozen=True)
class TextMessage:
    """
    Describes a text message in a chat.

    Parameters:
        role (Role, required):
            The role of the message.

        content (str, required):
            The content of the message.
    """

    role: Role
    content: str
    tool_calls: Optional[List[ToolCall]] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "TextMessage":
        tool_calls = json.get("tool_calls")
        return TextMessage(
            role=Role(json["role"]),
            content=json["content"],
            tool_calls=None
            if tool_calls is None
            else [ToolCall.from_json(tool_call) for tool_call in tool_calls],
        )

    # In multi-turn conversations the returned TextMessage is part of the chat
    # history and converted to the prompt. As such, it requires conversion to
    # json again. Here, the message content is a string, but can reuse the
    # _message_content_to_json function nonetheless.
    def to_json(self) -> Mapping[str, Any]:
        result = {
            "role": self.role.value,
            "content": _message_content_to_json(self.content),
        }
        if self.tool_calls is not None:
            result["tool_calls"] = [t.to_json() for t in self.tool_calls]
        return result


def _message_content_to_json(
    content: Union[str, List[Union[str, Image]]],
) -> Union[str, List[Mapping[str, Any]]]:
    if isinstance(content, str):
        return content
    else:
        result: List[Mapping[str, Any]] = []
        for chunk in content:
            if isinstance(chunk, str):
                result.append({"type": "text", "text": chunk})
            elif isinstance(chunk, Image):
                result.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": _image_to_data_uri(chunk)},
                    }
                )
            else:
                raise ValueError(
                    "The item in the prompt is not valid. Try either a string or an Image."
                )
        return result


def _image_to_data_uri(image: Image) -> str:
    buffer = BytesIO()
    image.save(buffer, format="PNG")
    base_64 = base64.b64encode(buffer.getvalue()).decode("utf-8")
    return f"data:image/png;base64,{base_64}"


@dataclass(frozen=True)
class StreamOptions:
    """
    Additional options to affect the streaming behavior.
    """

    # If set, an additional chunk will be streamed before the data: [DONE] message.
    # The usage field on this chunk shows the token usage statistics for the entire
    # request, and the choices field will always be an empty array.
    include_usage: bool


@dataclass(frozen=True)
class ToolFunction:
    type: Literal["function"]
    function: Any


@dataclass(frozen=True)
class ChatRequest:
    """
    Describes a chat request.

    Only supports a subset of the parameters of `CompletionRequest` for simplicity.
    See `CompletionRequest` for documentation on the parameters.
    """

    model: str
    messages: Sequence[Union[Message, TextMessage]]
    maximum_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    stream_options: Optional[StreamOptions] = None
    steering_concepts: Optional[List[str]] = None
    response_format: Optional[ResponseFormat] = None

    tools: Optional[List[Any]] = None
    tool_choice: Optional[Union[Literal["auto", "required", "none"], ToolFunction]] = (
        None
    )
    parallel_tool_calls: Optional[bool] = None

    def to_json(self) -> Mapping[str, Any]:
        payload = {k: v for k, v in asdict(self).items() if v is not None}
        payload["messages"] = [message.to_json() for message in self.messages]
        if self.response_format:
            # Handle Pydantic models by converting them to JSONSchema first
            if isinstance(self.response_format, type) and issubclass(
                self.response_format, BaseModel
            ):
                # This is a Pydantic model, convert it to JSONSchema
                from aleph_alpha_client.structured_output import JSONSchema

                json_schema = JSONSchema.from_pydantic(self.response_format)
                payload["response_format"] = json_schema.to_json()
            else:
                # This is already a JSONSchema or compatible object
                payload["response_format"] = self.response_format.to_json()
        return payload


class FinishReason(str, Enum):
    """
    The reason the model stopped generating tokens.

    This will be `stop` if the model hit a natural stop point or a provided stop
    sequence or length if the maximum number of tokens specified in the request
    was reached. If the API is unable to understand the stop reason emitted by
    one of the workers, content_filter is returned.
    """

    Stop = "stop"
    Length = "length"
    ContentFilter = "content_filter"
    ToolCalls = "tool_calls"


@dataclass(frozen=True)
class ChatResponse:
    """
    A simplified version of the chat response.

    As the `ChatRequest` does not support the `n` parameter (allowing for multiple return values),
    the `ChatResponse` assumes there to be only one choice.
    """

    finish_reason: FinishReason
    message: TextMessage

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "ChatResponse":
        first_choice = json["choices"][0]
        return ChatResponse(
            finish_reason=FinishReason(first_choice["finish_reason"]),
            message=TextMessage.from_json(first_choice["message"]),
        )


@dataclass(frozen=True)
class Usage:
    """
    Usage statistics for the completion request.

    When streaming is enabled, this field will be null by default.
    To include an additional usage-only message in the response stream, set stream_options.include_usage to true.
    """

    # Number of tokens in the generated completion.
    completion_tokens: int

    # Number of tokens in the prompt.
    prompt_tokens: int

    # Total number of tokens used in the request (prompt + completion).
    total_tokens: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "Usage":
        return Usage(
            completion_tokens=json["completion_tokens"],
            prompt_tokens=json["prompt_tokens"],
            total_tokens=json["total_tokens"],
        )


@dataclass(frozen=True)
class ChatStreamChunk:
    """
    A streamed chat completion chunk.

    Parameters:
        content (str, required):
            The content of the current chat completion. Will be empty for the first chunk of every completion stream and non-empty for the remaining chunks.

        role (Role, optional):
            The role of the current chat completion. Will be assistant for the first chunk of every completion stream and missing for the remaining chunks.
    """

    content: str
    role: Optional[Role]
    tool_calls: Optional[List[ToolCall]]

    @staticmethod
    def from_json(choice: Dict[str, Any]) -> "ChatStreamChunk":
        delta = choice["delta"]
        tool_calls = delta.get("tool_calls")
        return ChatStreamChunk(
            content=delta["content"],
            role=Role(delta.get("role")) if delta.get("role") else None,
            tool_calls=None
            if tool_calls is None
            else [ToolCall.from_json(tool_call) for tool_call in tool_calls],
        )


def stream_chat_item_from_json(
    json: Dict[str, Any],
) -> Union[ChatStreamChunk, FinishReason, Usage]:
    """Parse a chat event into one of the three possible types.

    This function takes two assumptions:

    1. If neither a finish reason nor a usage is present, the chunk contains a message.
    2. The chunk only carries information relevant to one of the three possible types, so no information is lost by choosing any one of them.
    """
    if (usage := json.get("usage")) is not None:
        return Usage.from_json(usage)

    first_choice = json["choices"][0]
    if (finish_reason := first_choice.get("finish_reason")) is not None:
        return FinishReason(finish_reason)

    return ChatStreamChunk.from_json(first_choice)


@dataclasses.dataclass
class ToolCallState:
    id: str
    arguments: str
    type: str
    function_name: str


async def process_chat_stream(
    s: AsyncGenerator[Dict[str, Any], None],
) -> AsyncGenerator[Union[ChatStreamChunk, Usage, ToolCall, FinishReason], None]:
    tool_calls: dict[str, ToolCallState] = {}
    # This loop collects all tool calls and emits them at the end of the stream *before* the finish reason
    # We have to do this because tool calls are streamed and distributed over different chunks
    async for json in s:
        if (usage := json.get("usage")) is not None:
            yield Usage.from_json(usage)
            continue

        first_choice = json["choices"][0]
        if (finish_reason := first_choice.get("finish_reason")) is not None:
            # Stream done, we have to yield all tool calls before the finish reason
            for tool in tool_calls.values():
                yield ToolCall(
                    id=tool.id,
                    type=tool.type,
                    function=FunctionCall(
                        name=tool.function_name,
                        arguments=tool.arguments,
                    ),
                )
            yield FinishReason(finish_reason)
            continue

        delta = first_choice.get("delta")
        if (tool_calls_json := delta.pop("tool_calls", None)) is not None:
            for tool_call in tool_calls_json:
                # We received a tool call delta, append all deltas to our state
                # or create the state if it is the first message for this tool call
                index = tool_call["index"]
                if index in tool_calls:
                    tool_calls[index].arguments += tool_call["function"]["arguments"]
                else:
                    function = tool_call["function"]
                    tool_calls[index] = ToolCallState(
                        id=tool_call["id"],
                        type=tool_call["type"],
                        function_name=function["name"],
                        arguments=function.get("arguments", ""),
                    )

            content = delta.get("content")
            if content is None or content == "":
                # We skip empty content here because our current api always returns content,
                # even for tool_call messages where this is not needed
                continue

        yield stream_chat_item_from_json(json)



from pathlib import Path
from aleph_alpha_client.explanation import (
    TargetScoreWithRaw,
    TextScoreWithRaw,
)
from aleph_alpha_client.prompt import Tokens
from aleph_alpha_client import (
    ControlTokenOverlap,
    ExplanationRequest,
    AsyncClient,
    Client,
    Image,
    Prompt,
    Text,
    CustomGranularity,
    TargetGranularity,
    PromptGranularity,
    ExplanationPostprocessing,
    ImageScore,
)
import pytest

# AsyncClient


@pytest.mark.vcr
async def test_can_explain_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = ExplanationRequest(
        prompt=Prompt(
            [
                Text.from_text(
                    "I am a programmer###I am French###I don't like pizza###My favourite food is"
                ),
                # " My favorite food is"
                Tokens.from_token_ids([4014, 36316, 5681, 387]),
            ]
        ),
        target=" pizza with cheese",
        prompt_granularity=CustomGranularity("###"),
        target_granularity=TargetGranularity.Token,
    )

    explanation = await async_client.explain(request, model=model_name)

    assert len(explanation.explanations) == 3
    assert all([len(exp.items) == 3 for exp in explanation.explanations])
    assert len(explanation.explanations[0].items[0].scores) == 4


# Client


@pytest.mark.vcr
def test_explanation_with_text_only(sync_client: Client, model_name: str):
    request = ExplanationRequest(
        prompt=Prompt.from_text("I am a programmer and French. My favourite food is"),
        target=" pizza with cheese",
        target_granularity=TargetGranularity.Token,
        normalize=True,
    )

    explanation = sync_client.explain(request, model=model_name)

    assert len(explanation.explanations) == 3
    assert all([len(exp.items) == 2 for exp in explanation.explanations])
    # At least one of the following options must be set in the request
    # to make all scores positive (or zero):
    # postprocessing=ExplanationPostProcessing.Absolute
    # postprocessing=ExplanationPostProcessing.Square
    # normalize=true
    for exp in explanation.explanations:
        for prompt_item in exp.items:
            assert all([score.score >= 0.0 for score in prompt_item.scores])


@pytest.mark.vcr
def test_explanation_with_multimodal_prompt(sync_client: Client, model_name: str):
    image_source_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    img = Image.from_image_source(image_source=str(image_source_path))

    request = ExplanationRequest(
        prompt=Prompt(
            [
                img,
                Text.from_text("I am a programmer and French. My favourite food is"),
                # " My favorite food is"
                Tokens.from_token_ids([4014, 36316, 5681, 387]),
            ]
        ),
        target=" pizza with cheese",
        prompt_granularity="sentence",
        postprocessing=ExplanationPostprocessing.Absolute,
        normalize=True,
        target_granularity=TargetGranularity.Token,
        control_token_overlap=ControlTokenOverlap.Complete,
    )

    explanation = sync_client.explain(request, model=model_name)

    assert len(explanation.explanations) == 3
    assert all([len(exp.items) == 4 for exp in explanation.explanations])
    # At least one of the following options must be set in the request
    # to make all scores positive (or zero):
    # postprocessing=ExplanationPostProcessing.Absolute
    # postprocessing=ExplanationPostProcessing.Square
    # normalize=true
    for exp in explanation.explanations:
        for prompt_item in exp.items:
            assert all([score.score >= 0.0 for score in prompt_item.scores])


@pytest.mark.vcr
def test_explanation_auto_granularity(sync_client: Client, model_name: str):
    image_source_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    img = Image.from_image_source(image_source=str(image_source_path))

    request = ExplanationRequest(
        prompt=Prompt(
            [
                img,
                Text.from_text("I am a programmer and French. My favourite food is"),
                # " My favorite food is"
                Tokens.from_token_ids([4014, 36316, 5681, 387]),
            ]
        ),
        target=" pizza with cheese",
        prompt_granularity=None,
    )

    explanation = sync_client.explain(request, model=model_name)

    assert len(explanation.explanations) == 1
    assert all([len(exp.items) == 4 for exp in explanation.explanations])


@pytest.mark.vcr
def test_explanation_of_image_in_pixels(sync_client: Client, model_name: str):
    image_source_path = Path(__file__).parent / "dog-and-cat-cover.jpg"
    img = Image.from_image_source(image_source=str(image_source_path))

    request = ExplanationRequest(
        prompt=Prompt(
            [
                img,
                Text.from_text("I am a programmer and French. My favourite food is"),
                # " My favorite food is"
                Tokens.from_token_ids([4014, 36316, 5681, 387]),
            ]
        ),
        target=" pizza with cheese",
        prompt_granularity=None,
    )

    explanation = sync_client.explain(request, model=model_name)

    explanation = explanation.with_image_prompt_items_in_pixels(request.prompt)
    assert len(explanation.explanations) == 1
    assert all([len(exp.items) == 4 for exp in explanation.explanations])
    assert all(
        [
            isinstance(image_score, ImageScore) and isinstance(image_score.left, int)
            for image_score in explanation.explanations[0].items[0].scores
        ]
    )


@pytest.mark.vcr
def test_explanation_with_text_from_request(sync_client: Client, model_name: str):
    request = ExplanationRequest(
        prompt=Prompt(
            [
                Text.from_text("I am a programmer and French. My favourite food is"),
                Tokens.from_token_ids([4014, 36316, 5681, 387]),
            ]
        ),
        target=" pizza with cheese",
        prompt_granularity=None,
        target_granularity=TargetGranularity.Token,
    )

    explanation = sync_client.explain(request, model=model_name)

    explanation = explanation.with_text_from_prompt(request)
    assert len(explanation.explanations) == 3
    assert all([len(exp.items) == 3 for exp in explanation.explanations])
    assert all(
        [
            isinstance(raw_text_score, TextScoreWithRaw)
            and isinstance(raw_text_score.text, str)
            for raw_text_score in explanation.explanations[0].items[0].scores
        ]
    )
    assert all(
        [
            isinstance(raw_text_score, TargetScoreWithRaw)
            and isinstance(raw_text_score.text, str)
            for raw_text_score in explanation.explanations[1].items[2].scores
        ]
    )


@pytest.mark.vcr
# Regression test for out-of-memory errors that could be triggered on the workers with the explanation job shown below
def test_explanation_with_token_granularities_oom_regression(sync_client: Client):
    prompt_text = """### Instruction:
Answer the question using the Source. If there's no answer, say "NO_ANSWER_IN_TEXT".

### Input:
Source: The Battle of Waterloo was fought on Sunday 18 June 1815, near Waterloo (at that time in the United Kingdom of the Netherlands, now in Belgium). A French army under the command of Napoleon was defeated by two of the armies of the Seventh Coalition. One of these was a British-led coalition consisting of units from the United Kingdom, the Netherlands, Hanover, Brunswick, and Nassau, under the command of the Duke of Wellington (referred to by many authors as the Anglo-allied army or Wellington's army). The other was composed of three corps of the Prussian army under the command of Field Marshal von Blücher (the fourth corps of this army fought at the Battle of Wavre on the same day). The battle marked the end of the Napoleonic Wars. The battle was contemporaneously known as the Battle of Mont Saint-Jean (France) or La Belle Alliance ("the Beautiful Alliance" – Prussia).
Question: Did India win or lost the Battle of Waterloo?

### Response:"""
    target_text = " India won the Battle of Waterloo."
    model_name = "luminous-base"
    request = ExplanationRequest(
        prompt=Prompt([Text(prompt_text, controls=[])]),
        target=target_text,
        prompt_granularity=PromptGranularity.Token,
        target_granularity=TargetGranularity.Token,
        control_factor=0.1,
        control_log_additive=True,
        postprocessing=None,
        control_token_overlap=ControlTokenOverlap.Partial,
        contextual_control_threshold=None,
    )
    explanation = sync_client.explain(request, model=model_name)
    assert len(explanation.explanations) == 7



import pytest
from aleph_alpha_client.completion import CompletionRequest
from aleph_alpha_client.steering import (
    SteeringConceptCreationRequest,
    SteeringPairedExample,
)

from .conftest import GenericClient, llama_prompt, requires_beta_features


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
@requires_beta_features
async def test_can_create_and_use_steering_concept(
    generic_client: GenericClient, chat_model_name: str
):
    # 1. We create a new steering concept with examples that steer the models
    # output away from formal english towards slang.
    response = await generic_client.create_steering_concept(
        create_sample_steering_concept_creation_request()
    )
    steering_concept_id = response.id
    assert isinstance(steering_concept_id, str) and len(steering_concept_id) > 0

    # 2. We ask the model to paraphrase "You are an honest man.", with and
    # without the steering concept applied.
    async def complete(**kwargs) -> str:
        response = await generic_client.complete(
            CompletionRequest(
                prompt=llama_prompt(
                    "Reply only with a paraphrased version of the following phrase: "
                    "You are an honest man."
                ),
                maximum_tokens=16,
                **kwargs,
            ),
            model=chat_model_name,
        )
        assert isinstance(response.completions[0].completion, str)
        return response.completions[0].completion

    base_completion = await complete()
    steered_completion = await complete(steering_concepts=[steering_concept_id])

    # The outputs for my test runs were:
    # base_completion == "Your integrity is evident in all that you do."
    # steered_completion == "You're a straight shooter, no games."
    #
    # In order to reduce the flakiness of tests we don't assert exact model
    # outputs.
    assert isinstance(base_completion, str)
    assert isinstance(steered_completion, str)
    assert base_completion != steered_completion


def create_sample_steering_concept_creation_request() -> SteeringConceptCreationRequest:
    return SteeringConceptCreationRequest(
        examples=[
            SteeringPairedExample(
                negative="I appreciate your valuable feedback on this matter.",
                positive="Thanks for the real talk, fam.",
            ),
            SteeringPairedExample(
                negative="The financial projections indicate significant growth potential.",
                positive="Yo, these numbers are looking mad stacked!",
            ),
        ]
    )



from aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.completion import (
    CompletionRequest,
    CompletionSummary,
    StreamChunk,
    StreamSummary,
)
from aleph_alpha_client.prompt import (
    ControlTokenOverlap,
    Image,
    Prompt,
    Text,
    TextControl,
    Tokens,
)
import pytest


@pytest.mark.vcr
async def test_can_complete_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = CompletionRequest(
        prompt=Prompt.from_text(""),
        maximum_tokens=7,
    )

    response = await async_client.complete(request, model=model_name)
    assert len(response.completions) == 1
    assert response.model_version is not None


@pytest.mark.vcr
async def test_can_use_streaming_support_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = CompletionRequest(
        prompt=Prompt.from_text(""),
        maximum_tokens=7,
    )

    stream_items = [
        stream_item
        async for stream_item in async_client.complete_with_streaming(
            request, model=model_name
        )
    ]

    assert len(stream_items) >= 3
    assert isinstance(stream_items[-3], StreamChunk)
    assert isinstance(stream_items[-2], StreamSummary)
    assert isinstance(stream_items[-1], CompletionSummary)


@pytest.mark.vcr
def test_complete_maximum_tokens_none(sync_client: Client, model_name: str):
    request = CompletionRequest(
        prompt=Prompt.from_text("Hello, World!"),
        maximum_tokens=7,
        stop_sequences=[","],
    )

    response = sync_client.complete(request, model=model_name)
    assert len(response.completions) == 1
    assert response.completions[0].completion is not None
    assert len(response.completions[0].completion) < 100


@pytest.mark.vcr
def test_complete(sync_client: Client, model_name: str):
    request = CompletionRequest(
        prompt=Prompt(
            [
                Text(
                    "Hello, World!",
                    controls=[
                        TextControl(start=1, length=5, factor=0.5),
                        TextControl(
                            start=1,
                            length=5,
                            factor=0.5,
                            token_overlap=ControlTokenOverlap.Complete,
                        ),
                    ],
                )
            ]
        ),
        maximum_tokens=7,
        tokens=False,
        log_probs=0,
        logit_bias={1: 2.0},
    )

    response = sync_client.complete(request, model=model_name)

    assert len(response.completions) == 1
    assert response.model_version is not None


@pytest.mark.vcr
def test_complete_with_token_ids(sync_client: Client, model_name: str):
    request = CompletionRequest(
        prompt=Prompt.from_tokens([49222, 2998]),  # Hello world
        maximum_tokens=32,
    )

    response = sync_client.complete(request, model=model_name)

    assert len(response.completions) == 1
    assert response.model_version is not None


@pytest.mark.vcr
def test_complete_with_optimized_prompt(
    sync_client: Client, model_name: str, prompt_image: Image
):
    prompt_tokens = Tokens.from_token_ids([1, 2])
    request = CompletionRequest(
        prompt=Prompt(
            [
                Text.from_text(" Hello "),
                Text.from_text(" world! "),
                prompt_image,
                prompt_tokens,
                Text.from_text(" The "),
                Text.from_text(" end "),
            ]
        ),
        maximum_tokens=5,
    )

    response = sync_client.complete(request, model=model_name)

    assert response.optimized_prompt is not None
    assert response.optimized_prompt.items[0] == Text.from_text("Hello  world! ")
    assert response.optimized_prompt.items[2] == prompt_tokens
    assert response.optimized_prompt.items[3] == Text.from_text(" The  end")
    assert isinstance(response.optimized_prompt.items[1], Image)


@pytest.mark.vcr
def test_complete_with_echo(sync_client: Client, model_name: str, prompt_image: Image):
    request = CompletionRequest(
        prompt=Prompt.from_text("Hello world"),
        maximum_tokens=0,
        tokens=True,
        echo=True,
        log_probs=0,
    )

    response = sync_client.complete(request, model=model_name)
    completion_result = response.completions[0]
    assert completion_result.completion == " Hello world"
    assert completion_result.completion_tokens is not None
    assert len(completion_result.completion_tokens) > 0
    assert completion_result.log_probs is not None
    assert len(completion_result.log_probs) > 0


@pytest.mark.vcr
def test_num_tokens_prompt_total_with_best_of(sync_client: Client, model_name: str):
    tokens = [49222, 2998]  # Hello world
    best_of = 2
    request = CompletionRequest(
        prompt=Prompt.from_tokens(tokens),
        best_of=best_of,
        maximum_tokens=1,
    )

    response = sync_client.complete(request, model=model_name)
    assert response.num_tokens_prompt_total == len(tokens) * best_of


@pytest.mark.vcr
def test_num_tokens_generated_with_best_of(sync_client: Client, model_name: str):
    hello_world = [49222, 2998]  # Hello world
    best_of = 2
    request = CompletionRequest(
        prompt=Prompt.from_tokens(hello_world),
        best_of=best_of,
        maximum_tokens=1,
        tokens=True,
    )

    response = sync_client.complete(request, model=model_name)
    completion_result = response.completions[0]
    assert completion_result.completion_tokens is not None
    number_tokens_completion = len(completion_result.completion_tokens)

    assert response.num_tokens_generated == best_of * number_tokens_completion



import pytest

from aleph_alpha_client.reranking import RerankRequest, RerankResponse
from tests.conftest import GenericClient


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
async def test_can_rerank_documents(
    generic_client: GenericClient, rerank_model_name: str
):
    request = RerankRequest(
        query="What is the capital of France?",
        documents=[
            "The capital of Brazil is Brasilia.",
            "The capital of France is Paris.",
            "Horses and cows are both animals.",
        ],
    )

    response = await generic_client.rerank(request, model=rerank_model_name)
    assert isinstance(response, RerankResponse)
    assert len(response.results) == 3
    for result in response.results:
        assert result.index >= 0
        assert result.index < 3
        assert isinstance(result.relevance_score, float)

    # The document about Paris should have the highest relevance score
    highest_result = response.results[0]
    assert highest_result.relevance_score > 0
    assert highest_result.index == 1

    assert response.usage.total_tokens > 0


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
async def test_can_rerank_with_top_n(
    generic_client: GenericClient, rerank_model_name: str
):
    request = RerankRequest(
        query="What is the capital of France?",
        documents=[
            "The capital of Brazil is Brasilia.",
            "The capital of France is Paris.",
            "Horses and cows are both animals.",
        ],
        top_n=2,
    )

    response = await generic_client.rerank(request, model=rerank_model_name)
    assert isinstance(response, RerankResponse)
    assert len(response.results) == 2


@pytest.mark.vcr
@pytest.mark.parametrize(
    "generic_client", ["sync_client", "async_client"], indirect=True
)
async def test_rerank_returns_results_ordered_by_relevance(
    generic_client: GenericClient, rerank_model_name: str
):
    request = RerankRequest(
        query="What is the capital of France?",
        documents=[
            "The capital of Brazil is Brasilia.",
            "The capital of France is Paris.",
            "Horses and cows are both animals.",
        ],
    )

    response = await generic_client.rerank(request, model=rerank_model_name)

    # Verify results are sorted by relevance score in descending order
    scores = [r.relevance_score for r in response.results]
    assert scores == sorted(scores, reverse=True)



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Mapping, Optional, Sequence, Union

from aleph_alpha_client.prompt import Prompt


@dataclass(frozen=True)
class CompletionRequest:
    """
    Describes a completion request

    Parameters:
        prompt:
            The text or image prompt to be completed.
            Unconditional completion can be started with an empty string (default).
            The prompt may contain a zero shot or few shot task.

        maximum_tokens (int, optional, default None):
            The maximum number of tokens to be generated.
            Completion will terminate after the maximum number of tokens is reached. Increase this value to generate longer texts.
            A text is split into tokens. Usually there are more tokens than words.
            The maximum supported number of tokens depends on the model (for luminous-base, it may not exceed 2048 tokens).
            The prompt's tokens plus the maximum_tokens request must not exceed this number. If set to None, the model will stop
            generating tokens either if it outputs a sequence specified in `stop_sequences` or if it reaches its technical limit.
            For most models, this means that the sum of input and output tokens is equal to its context window.

        temperature (float, optional, default 0.0)
            A higher sampling temperature encourages the model to produce less probable outputs ("be more creative"). Values are expected in a range from 0.0 to 1.0. Try high values (e.g. 0.9) for a more "creative" response and the default 0.0 for a well defined and repeatable answer.

            It is recommended to use either temperature, top_k or top_p and not all at the same time. If a combination of temperature, top_k or top_p is used rescaling of logits with temperature will be performed first. Then top_k is applied. Top_p follows last.

        top_k (int, optional, default 0)
            Introduces random sampling from generated tokens by randomly selecting the next token from the k most likely options. A value larger than 1 encourages the model to be more creative. Set to 0 if repeatable output is to be produced.
            It is recommended to use either temperature, top_k or top_p and not all at the same time. If a combination of temperature, top_k or top_p is used rescaling of logits with temperature will be performed first. Then top_k is applied. Top_p follows last.

        top_p (float, optional, default 0.0)
            Introduces random sampling for generated tokens by randomly selecting the next token from the smallest possible set of tokens whose cumulative probability exceeds the probability top_p. Set to 0.0 if repeatable output is to be produced.
            It is recommended to use either temperature, top_k or top_p and not all at the same time. If a combination of temperature, top_k or top_p is used rescaling of logits with temperature will be performed first. Then top_k is applied. Top_p follows last.

        presence_penalty (float, optional, default 0.0)
            The presence penalty reduces the likelihood of generating tokens that are already present in the
            generated text (`repetition_penalties_include_completion=true`) respectively the prompt (`repetition_penalties_include_prompt=true`).
            Presence penalty is independent of the number of occurrences. Increase the value to produce text that is not repeating the input.

        frequency_penalty (float, optional, default 0.0)
            The frequency penalty reduces the likelihood of generating tokens that are already present in the
            generated text (`repetition_penalties_include_completion=true`) respectively the prompt (`repetition_penalties_include_prompt=true`).
            Frequency penalty is dependent on the number of occurrences of a token.

        repetition_penalties_include_prompt (bool, optional, default False)
            Flag deciding whether presence penalty or frequency penalty are updated from the prompt

        use_multiplicative_presence_penalty (bool, optional, default True)
            Flag deciding whether presence penalty is applied multiplicatively (True) or additively (False). This changes the formula stated for presence and frequency penalty.

        penalty_bias (string, optional)
            If set, all tokens in this text will be used in addition to the already penalized tokens for repetition penalties.
            These consist of the already generated completion tokens if ``repetition_penalties_include_completion`` is set to ``true``
            and the prompt tokens, if ``repetition_penalties_include_prompt`` is set to ``true``,

            *Potential use case for a chatbot-based completion:*

            Instead of using ``repetition_penalties_include_prompt``, construct a new string with only the chatbot's responses included. You would leave out any tokens you use for stop sequences (i.e. ``\\nChatbot:``), and all user messages.

            With this bias, if you turn up the repetition penalties, you can avoid having your chatbot repeat itself, but not penalize the chatbot from mirroring language provided by the user.

        penalty_exceptions (List(str), optional)
            List of strings that may be generated without penalty, regardless of other penalty settings.

            This is particularly useful for any completion that uses a structured few-shot prompt. For example, if you have a prompt such as:

            ::

                I want to travel to a location, where I can enjoy both beaches and mountains.

                - Lake Garda, Italy. This large Italian lake in the southern alps features gravel beaches and mountainside hiking trails.
                - Mallorca, Spain. This island is famous for its sandy beaches, turquoise water and hilly landscape.
                - Lake Tahoe, California. This famous lake in the Sierra Nevada mountains offers an amazing variety of outdoor activities.
                -

            You could set ``penalty_exceptions`` to ``["\\n-"]`` to not penalize the generation of a new list item, but still increase other penalty settings to encourage the generation of new list items without repeating itself.

            By default, we will also include any ``stop_sequences`` you have set, since completion performance can be degraded if expected stop sequences are penalized. You can disable this behavior by settings ``penalty_exceptions_include_stop_sequences`` to ``false``.

        penalty_exceptions_include_stop_sequences (bool, optional, default true)
            By default, we include any ``stop_sequences`` in ``penalty_exceptions``, to not penalize the presence of stop sequences that are present in few-shot prompts to provide structure to your completions.

            You can set this to ``false`` if you do not want this behavior.

            See the description of ``penalty_exceptions`` above for more information on what ``penalty_exceptions`` are used for.

        best_of (int, optional, default None)
            Generates best_of completions server-side and returns the "best" (the one with the highest log probability per token). Results cannot be streamed.
            When used with n, best_of controls the number of candidate completions and n specifies how many to return – best_of must be greater than n.

        n (int, optional, default 1)
            How many completions to generate for each prompt.

        logit_bias (dict mapping token ids to score, optional, default None)
            The logit bias allows to influence the likelihood of generating tokens. A dictionary mapping token ids (int) to a bias (float) can be provided. Such bias is added to the logits as generated by the model.

        log_probs (int, optional, default None)
            Number of top log probabilities to be returned for each generated token. Log probabilities may be used in downstream tasks or to assess the model's certainty when producing tokens.

            If set to 0, you will always get the log probability of the sampled token. 1 or more will return the argmax token(s) plus the sampled one, if not already included.

        stop_sequences (List(str), optional, default None)
            List of strings which will stop generation if they're generated. Stop sequences may be helpful in structured texts.

            Example: In a question answering scenario a text may consist of lines starting with either "Question: " or "Answer: " (alternating). After producing an answer, the model will be likely to generate "Question: ". "Question: " may therefore be used as stop sequence in order not to have the model generate more questions but rather restrict text generation to the answers.

        tokens (bool, optional, default False)
            return tokens of completion

        disable_optimizations  (bool, optional, default False)
            We continually research optimal ways to work with our models. By default, we apply these optimizations to both your prompt and  completion for you.

            Our goal is to improve your results while using our API. But you can always pass disable_optimizations: true and we will leave your prompt and completion untouched.

        minimum_tokens (int, default 0)
            Generate at least this number of tokens before an end-of-text token is generated.

        echo (bool, default False)
            Echo the prompt in the completion. This may be especially helpful when log_probs is set to return logprobs for the prompt.

        use_multiplicative_frequency_penalty (bool, default False)
            Flag deciding whether frequency penalty is applied multiplicatively (True) or additively (False).

        sequence_penalty (float, default 0.0)
            Increasing the sequence penalty reduces the likelihood of reproducing token sequences that already appear in the prompt
            (if repetition_penalties_include_prompt is True) and prior completion (if repetition_penalties_include_completion is True).

        sequence_penalty_min_length (int, default 2)
            Minimal number of tokens to be considered as sequence. Must be greater or equal 2.

        use_multiplicative_sequence_penalty (bool, default False)
            Flag deciding whether sequence penalty is applied multiplicatively (True) or additively (False).

        completion_bias_inclusion (List[str], default [])
            Bias the completion to only generate options within this list;
            all other tokens are disregarded at sampling

            Note that strings in the inclusion list must not be prefixes
            of strings in the exclusion list and vice versa

        completion_bias_inclusion_first_token_only (bool, default False)
            Only consider the first token for the completion_bias_inclusion

        completion_bias_exclusion (List[str], default [])
            Bias the completion to NOT generate options within this list;
            all other tokens are unaffected in sampling

            Note that strings in the inclusion list must not be prefixes
            of strings in the exclusion list and vice versa

        completion_bias_exclusion_first_token_only (bool, default False)
            Only consider the first token for the completion_bias_exclusion

        contextual_control_threshold (float, default None)
            If set to None, attention control parameters only apply to those tokens that have
            explicitly been set in the request.
            If set to a non-None value, we apply the control parameters to similar tokens as well.
            Controls that have been applied to one token will then be applied to all other tokens
            that have at least the similarity score defined by this parameter.
            The similarity score is the cosine similarity of token embeddings.

        control_log_additive (bool, default True)
            True: apply control by adding the log(control_factor) to attention scores.
            False: apply control by (attention_scores - - attention_scores.min(-1)) * control_factor

        repetition_penalties_include_completion (bool, optional, default True)
            Flag deciding whether presence penalty or frequency penalty are updated from the completion

        raw_completion (bool, default False)
            Setting this parameter to true forces the raw completion of the model to be returned.
            For some models, we may optimize the completion that was generated by the model and
            return the optimized completion in the completion field of the CompletionResponse.
            The raw completion, if returned, will contain the un-optimized completion.

        steering_concepts (Optional[list[str]], default None)
            Names of the steering vectors to apply on this task. This steers the output in the
            direction given by positive examples, and away from negative examples if provided.

    Examples:
        >>> prompt = Prompt.from_text("Provide a short description of AI:")
        >>> request = CompletionRequest(prompt=prompt, maximum_tokens=20)
    """

    prompt: Prompt
    maximum_tokens: Optional[int] = None
    temperature: float = 0.0
    top_k: int = 0
    top_p: float = 0.0
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    repetition_penalties_include_prompt: bool = False
    use_multiplicative_presence_penalty: bool = False
    penalty_bias: Optional[str] = None
    penalty_exceptions: Optional[List[str]] = None
    penalty_exceptions_include_stop_sequences: Optional[bool] = None
    best_of: Optional[int] = None
    n: int = 1
    logit_bias: Optional[Dict[int, float]] = None
    log_probs: Optional[int] = None
    stop_sequences: Optional[List[str]] = None
    tokens: bool = False
    disable_optimizations: bool = False
    minimum_tokens: int = 0
    echo: bool = False
    use_multiplicative_frequency_penalty: bool = False
    sequence_penalty: float = 0.0
    sequence_penalty_min_length: int = 2
    use_multiplicative_sequence_penalty: bool = False
    completion_bias_inclusion: Optional[Sequence[str]] = None
    completion_bias_inclusion_first_token_only: bool = False
    completion_bias_exclusion: Optional[Sequence[str]] = None
    completion_bias_exclusion_first_token_only: bool = False
    contextual_control_threshold: Optional[float] = None
    control_log_additive: Optional[bool] = True
    repetition_penalties_include_completion: bool = True
    raw_completion: bool = False
    steering_concepts: Optional[List[str]] = None

    def to_json(self) -> Mapping[str, Any]:
        payload = {k: v for k, v in self._asdict().items() if v is not None}
        payload["prompt"] = self.prompt.to_json()
        return payload

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class CompletionResult:
    log_probs: Optional[Sequence[Mapping[str, Optional[float]]]] = None
    completion: Optional[str] = None
    completion_tokens: Optional[Sequence[str]] = None
    finish_reason: Optional[str] = None
    raw_completion: Optional[str] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "CompletionResult":
        return CompletionResult(
            log_probs=json.get("log_probs"),
            completion=json.get("completion"),
            completion_tokens=json.get("completion_tokens"),
            finish_reason=json.get("finish_reason"),
            raw_completion=json.get("raw_completion"),
        )

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class CompletionResponse:
    """
    Describes a completion response

    Parameters:
        model_version:
            Model name and version (if any) of the used model for inference.
        completions:
            List of completions; may contain only one entry if no more are requested (see parameter n).
        num_tokens_prompt_total:
            Number of tokens combined across all completion tasks.
            In particular, if you set best_of or n to a number larger than 1 then we report the
            combined prompt token count for all best_of or n tasks.
        num_tokens_generated:
            Number of tokens combined across all completion tasks.
            If multiple completions are returned or best_of is set to a value greater than 1 then
            this value contains the combined generated token count.
        optimized_prompt:
            Describes prompt after optimizations. This field is only returned if the flag
            `disable_optimizations` flag is not set and the prompt has actually changed.
    """

    model_version: str
    completions: Sequence[CompletionResult]
    num_tokens_prompt_total: int
    num_tokens_generated: int
    optimized_prompt: Optional[Prompt] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "CompletionResponse":
        optimized_prompt_json = json.get("optimized_prompt")
        return CompletionResponse(
            model_version=json["model_version"],
            completions=[
                CompletionResult.from_json(item) for item in json["completions"]
            ],
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
            num_tokens_generated=json["num_tokens_generated"],
            optimized_prompt=(
                Prompt.from_json(optimized_prompt_json)
                if optimized_prompt_json
                else None
            ),
        )

    def to_json(self) -> Mapping[str, Any]:
        return {
            **self._asdict(),
            "completions": [completion._asdict() for completion in self.completions],
        }

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


CompletionResponseStreamItem = Union[
    "StreamChunk", "StreamSummary", "CompletionSummary"
]


def stream_item_from_json(json: Dict[str, Any]) -> CompletionResponseStreamItem:
    if json["type"] == "stream_chunk":
        return StreamChunk.from_json(json)
    elif json["type"] == "stream_summary":
        return StreamSummary.from_json(json)
    elif json["type"] == "completion_summary":
        return CompletionSummary.from_json(json)
    else:
        raise ValueError(f"Unknown stream item type: {json['type']}")


@dataclass(frozen=True)
class StreamChunk:
    """
    Describes a chunk of a completion stream

    Parameters:
        index:
            The index of the stream that this chunk belongs to.
            This is relevant if multiple completion streams are requested (see parameter n).
        log_probs:
            The log probabilities of the generated tokens.
        completion:
            The generated tokens formatted as single a string.
        raw_completion:
            The generated tokens including special tokens formatted as single a string.
        completion_tokens:
            The generated tokens as a list of strings.
    """

    index: int
    log_probs: Optional[Sequence[Mapping[str, Optional[float]]]]
    completion: str
    raw_completion: Optional[str]
    completion_tokens: Optional[Sequence[str]]

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "StreamChunk":
        return StreamChunk(
            index=json["index"],
            log_probs=json.get("log_probs"),
            completion=json["completion"],
            raw_completion=json.get("raw_completion"),
            completion_tokens=json.get("completion_tokens"),
        )

    def to_json(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class StreamSummary:
    """
    Denotes the end of a completion stream

    Parameters:
        index:
            The index of the stream that is being terminated.
            This is relevant if multiple completion streams are requested (see parameter n).
        model_version:
            Model name and version (if any) of the used model for inference.
        finish_reason:
            The reason why the model stopped generating new tokens.
    """

    index: int
    model_version: str
    finish_reason: str

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "StreamSummary":
        return StreamSummary(
            index=json["index"],
            model_version=json["model_version"],
            finish_reason=json["finish_reason"],
        )

    def to_json(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class CompletionSummary:
    """
    Denotes the end of all completion streams

    Parameters:
        optimized_prompt:
            Describes prompt after optimizations. This field is only returned if the flag
            `disable_optimizations` flag is not set and the prompt has actually changed.
        num_tokens_prompt_total:
            Number of tokens combined across all completion tasks.
            In particular, if you set best_of or n to a number larger than 1 then we report the
            combined prompt token count for all best_of or n tasks.
        num_tokens_generated:
            Number of tokens combined across all completion tasks.
            If multiple completions are returned or best_of is set to a value greater than 1 then
            this value contains the combined generated token count.
    """

    optimized_prompt: Optional[Prompt]
    num_tokens_prompt_total: int
    num_tokens_generated: int

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "CompletionSummary":
        optimized_prompt_json = json.get("optimized_prompt")
        return CompletionSummary(
            optimized_prompt=(
                Prompt.from_json(optimized_prompt_json)
                if optimized_prompt_json
                else None
            ),
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
            num_tokens_generated=json["num_tokens_generated"],
        )



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class TranslationRequest:
    """Request for translation.

    Parameters:
        model (string, required):
            The name of the model to be used for the translation.

        source (string, required):
            The input text to be translated.

        target_language (string, required):
            The desired target language into which the input text should be translated. The language
            must be specified using ISO 639 (and RFC 1766) language codes such as "en" for English,
            "de" for German, "fr" for French, etc. For a list of supported languages, refer to the
            `/languages` endpoint.
    """

    model: str
    source: str
    target_language: str

    def to_json(self) -> dict:
        """Convert the request to a JSON-serializable dictionary."""
        return {
            "model": self.model,
            "source": self.source,
            "target_language": self.target_language,
        }


@dataclass
class TranslationSegment:
    """A segment of translated text with its source and quality score.

    Parameters:
        source (string):
            The input text to be translated.
        translation (string):
            The translated output text of the segment.
        score (float):
            Estimate for the overall quality of the translation on a scale of 0 to 1.
    """

    source: str
    translation: str
    score: float


@dataclass
class TranslationResponse:
    """Response from a translation request.

    Parameters:
        translation (string):
            The complete translated output text
        score (float):
            Estimate for the overall quality of the translation on a scale of 0 to 1.
        segments (Optional[List[TranslationSegment]]):
            List of translated segments. May be None if no segments are returned.
        num_tokens_prompt_total (int):
            Total number of tokens in the prompt. May be zero as the current translation
            models don't expose tokens.
        num_tokens_generated (int):
            Total number of tokens generated. May be zero as the current translation
            models don't expose tokens.
    """

    translation: str
    score: float
    segments: Optional[List[TranslationSegment]]
    num_tokens_prompt_total: int
    num_tokens_generated: int

    @classmethod
    def from_json(cls, json: dict) -> "TranslationResponse":
        """Create a TranslationResponse from a JSON dictionary."""
        segments = None
        if json.get("segments") is not None:
            segments = [
                TranslationSegment(
                    source=segment["source"],
                    translation=segment["translation"],
                    score=segment["score"],
                )
                for segment in json["segments"]
            ]

        return cls(
            translation=json["translation"],
            score=json["score"],
            segments=segments,
            num_tokens_prompt_total=json["num_tokens_prompt_total"],
            num_tokens_generated=json["num_tokens_generated"],
        )



from dataclasses import asdict, dataclass
from typing import Any, Dict, Mapping, Optional, Sequence


@dataclass(frozen=True)
class TokenizationRequest:
    """Describes a tokenization request.

    Parameters
        prompt (str)
            The text prompt which should be converted into tokens

        tokens (bool)
            True to extract text-tokens

        token_ids (bool)
            True to extract token-ids

    Returns
        TokenizationResponse

    Examples:
        >>> request = TokenizationRequest(prompt="This is an example.", tokens=True, token_ids=True)
    """

    prompt: str
    tokens: bool
    token_ids: bool

    def to_json(self) -> Mapping[str, Any]:
        return self._asdict()

    def _asdict(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass(frozen=True)
class TokenizationResponse:
    tokens: Optional[Sequence[str]] = None
    token_ids: Optional[Sequence[int]] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "TokenizationResponse":
        return TokenizationResponse(
            tokens=json.get("tokens"), token_ids=json.get("token_ids")
        )



from aleph_alpha_client.aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.evaluation import EvaluationRequest
from aleph_alpha_client.prompt import Prompt
import pytest

# AsyncClient


@pytest.mark.vcr
async def test_can_evaluate_with_async_client(
    async_client: AsyncClient, model_name: str
):
    request = EvaluationRequest(
        prompt=Prompt.from_text("hello"), completion_expected="world"
    )

    response = await async_client.evaluate(request, model=model_name)
    assert response.model_version is not None
    assert response.result is not None
    assert response.num_tokens_prompt_total >= 1


# Client


@pytest.mark.vcr
def test_evaluate(sync_client: Client, model_name: str):
    request = EvaluationRequest(
        prompt=Prompt.from_text("hello"), completion_expected="world"
    )

    response = sync_client.evaluate(request, model=model_name)

    assert response.model_version is not None
    assert response.result is not None
    assert response.num_tokens_prompt_total >= 1



from aleph_alpha_client import (
    ControlTokenOverlap,
    Prompt,
    TokenControl,
    Image,
    ImageControl,
)
from aleph_alpha_client.aleph_alpha_client import Client
from aleph_alpha_client.completion import CompletionRequest
from aleph_alpha_client.prompt import Text, TextControl
import pytest


def test_init_prompt_with_str():
    text = "text prompt"
    prompt = Prompt(text)

    assert prompt == Prompt.from_text(text)


def test_serialize_token_ids():
    tokens = [1, 2, 3, 4]
    prompt = Prompt.from_tokens(tokens)
    serialized_prompt = prompt.to_json()

    assert serialized_prompt == [
        {"type": "token_ids", "data": [1, 2, 3, 4], "controls": []}
    ]


def test_serialize_token_ids_with_controls():
    tokens = [1, 2, 3, 4]
    prompt = Prompt.from_tokens(
        tokens,
        controls=[
            TokenControl(pos=0, factor=0.25),
            TokenControl(pos=1, factor=0.5),
        ],
    )
    serialized_prompt = prompt.to_json()

    assert serialized_prompt == [
        {
            "type": "token_ids",
            "data": tokens,
            "controls": [{"index": 0, "factor": 0.25}, {"index": 1, "factor": 0.5}],
        }
    ]


def test_serialize_text_with_controls():
    prompt_text = "An apple a day"
    prompt = Prompt.from_text(
        prompt_text,
        [
            TextControl(start=3, length=5, factor=1.5),
            TextControl(
                start=3,
                length=5,
                factor=1.5,
                token_overlap=ControlTokenOverlap.Complete,
            ),
        ],
    )

    serialized_prompt = prompt.to_json()

    assert serialized_prompt == [
        {
            "type": "text",
            "data": prompt_text,
            "controls": [
                {"start": 3, "length": 5, "factor": 1.5},
                {"start": 3, "length": 5, "factor": 1.5, "token_overlap": "complete"},
            ],
        }
    ]


def test_serialize_image_with_controls():
    image = Image.from_file(
        "tests/dog-and-cat-cover.jpg",
        [
            ImageControl(0.0, 0.0, 0.5, 0.5, 0.5),
            ImageControl(
                left=0.0,
                top=0.0,
                width=0.5,
                height=0.5,
                factor=0.5,
                token_overlap=ControlTokenOverlap.Partial,
            ),
        ],
    )
    prompt = Prompt.from_image(image)
    serialized_prompt = prompt.to_json()

    assert serialized_prompt == [
        {
            "type": "image",
            "data": image.base_64,
            "controls": [
                {
                    "rect": {
                        "left": 0.0,
                        "top": 0.0,
                        "width": 0.5,
                        "height": 0.5,
                    },
                    "factor": 0.5,
                },
                {
                    "rect": {
                        "left": 0.0,
                        "top": 0.0,
                        "width": 0.5,
                        "height": 0.5,
                    },
                    "factor": 0.5,
                    "token_overlap": "partial",
                },
            ],
        }
    ]


@pytest.mark.vcr
def test_image_controls_with_cats_and_dogs(sync_client: Client):
    base_completion = _complete_cats_and_dogs(sync_client)
    completion_with_suppressed_cat = _complete_cats_and_dogs(
        sync_client,
        controls=[
            # Suppress the cat
            ImageControl(left=0.5, top=0.0, width=0.25, height=1.0, factor=0.0)
        ],
    )
    assert " dog" in base_completion
    assert " dog" in completion_with_suppressed_cat
    assert " cat" in base_completion
    assert " cat" not in completion_with_suppressed_cat


def _complete_cats_and_dogs(client: Client, **kwargs) -> str:
    image = Image.from_file_with_cropping(
        "tests/dog-and-cat-cover.jpg",
        # crop exactly 600x600 pixels out of the image
        300,
        0,
        600,
        **kwargs,
    )
    text = Text.from_text("A picture of ")
    request = CompletionRequest(
        prompt=Prompt([image, text]),
        maximum_tokens=16,
        control_log_additive=True,
        disable_optimizations=False,
    )
    result = client.complete(request, model="luminous-base")
    completion = result.completions[0].completion
    assert isinstance(completion, str)
    return completion



from pytest_httpserver import HTTPServer
import os
import pytest

from aleph_alpha_client.version import MIN_API_VERSION
from aleph_alpha_client.aleph_alpha_client import AsyncClient, Client
from aleph_alpha_client.completion import (
    CompletionRequest,
    CompletionResponse,
    CompletionResult,
)
from aleph_alpha_client.prompt import Prompt


def test_api_version_mismatch_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data("0.0.0")

    with pytest.raises(RuntimeError):
        Client(host=httpserver.url_for(""), token="TEST_TOKEN").validate_version()


async def test_api_version_mismatch_async_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data("0.0.0")

    with pytest.raises(RuntimeError):
        async with AsyncClient(
            host=httpserver.url_for(""), token="TEST_TOKEN"
        ) as client:
            await client.validate_version()


def test_api_version_correct_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)
    Client(host=httpserver.url_for(""), token="TEST_TOKEN").validate_version()


async def test_api_version_correct_async_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)
    async with AsyncClient(host=httpserver.url_for(""), token="TEST_TOKEN") as client:
        await client.validate_version()


async def test_can_use_async_client_without_context_manager(model_name: str):
    request = CompletionRequest(
        prompt=Prompt.from_text(""),
        maximum_tokens=7,
    )
    token = os.environ["TEST_TOKEN"]
    client = AsyncClient(token, host=os.environ["TEST_API_URL"])
    try:
        _ = await client.complete(request, model=model_name)
    finally:
        await client.close()


def test_nice_flag_on_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)

    httpserver.expect_request(
        "/complete", query_string={"nice": "true"}
    ).respond_with_json(
        CompletionResponse(
            "model_version",
            [
                CompletionResult(
                    log_probs=[],
                    completion="foo",
                )
            ],
            num_tokens_prompt_total=2,
            num_tokens_generated=1,
        ).to_json()
    )

    client = Client(host=httpserver.url_for(""), token="TEST_TOKEN", nice=True)

    request = CompletionRequest(prompt=Prompt.from_text("Hello world"))
    client.complete(request, model="luminous")


async def test_nice_flag_on_async_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)

    httpserver.expect_request(
        "/complete",
        query_string={"nice": "true"},
    ).respond_with_json(
        CompletionResponse(
            "model_version",
            [CompletionResult(log_probs=[], completion="foo")],
            num_tokens_prompt_total=2,
            num_tokens_generated=1,
        ).to_json()
    )

    request = CompletionRequest(prompt=Prompt.from_text("Hello world"))

    async with AsyncClient(
        host=httpserver.url_for(""),
        token="TEST_TOKEN",
        nice=True,
        request_timeout_seconds=1,
    ) as client:
        await client.complete(request, model="luminous")


def test_tags_on_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)

    request = CompletionRequest(prompt=Prompt.from_text("Hello world"))
    body = {k: v for k, v in request.to_json().items() if v is not None}
    body["tags"] = ["tim-tagger"]
    body["model"] = "luminous"
    httpserver.expect_request("/complete", json=body).respond_with_json(
        CompletionResponse(
            "model_version",
            [
                CompletionResult(
                    log_probs=[],
                    completion="foo",
                )
            ],
            num_tokens_prompt_total=2,
            num_tokens_generated=1,
        ).to_json()
    )

    client = Client(
        host=httpserver.url_for(""),
        request_timeout_seconds=1,
        token="TEST_TOKEN",
        tags=["tim-tagger"],
    )

    client.complete(request, model="luminous")


async def test_tags_on_async_client(httpserver: HTTPServer):
    httpserver.expect_request("/version").respond_with_data(MIN_API_VERSION)

    request = CompletionRequest(prompt=Prompt.from_text("Hello world"))
    body = {k: v for k, v in request.to_json().items() if v is not None}
    body["tags"] = ["tim-tagger"]
    body["model"] = "luminous"
    httpserver.expect_request("/complete", json=body).respond_with_json(
        CompletionResponse(
            "model_version",
            [CompletionResult(log_probs=[], completion="foo")],
            num_tokens_prompt_total=2,
            num_tokens_generated=1,
        ).to_json()
    )

    async with AsyncClient(
        host=httpserver.url_for(""), token="TEST_TOKEN", tags=["tim-tagger"]
    ) as client:
        await client.complete(request, model="luminous")


def test_available_models_sync_client(sync_client: Client, model_name: str):
    models = sync_client.models()
    assert model_name in {model["name"] for model in models}


@pytest.mark.vcr
async def test_available_models_async_client(
    async_client: AsyncClient, model_name: str
):
    models = await async_client.models()
    assert model_name in {model["name"] for model in models}



from .prompt import (
    ControlTokenOverlap,
    Image,
    ImageControl,
    Prompt,
    Text,
    TextControl,
    TokenControl,
    Tokens,
)
from .prompt_template import PromptTemplate
from .aleph_alpha_client import (
    POOLING_OPTIONS,
    AsyncClient,
    BusyError,
    Client,
    QuotaError,
)
from .completion import CompletionRequest, CompletionResponse
from .chat import ChatRequest, Message, ChatResponse
from .detokenization import DetokenizationRequest, DetokenizationResponse
from .document import Document
from .embedding import (
    EmbeddingRequest,
    EmbeddingV2Request,
    EmbeddingResponse,
    EmbeddingV2Response,
    BatchSemanticEmbeddingRequest,
    BatchSemanticEmbeddingResponse,
    SemanticEmbeddingRequest,
    SemanticEmbeddingResponse,
    SemanticRepresentation,
)
from .evaluation import EvaluationRequest, EvaluationResponse
from .explanation import (
    CustomGranularity,
    Explanation,
    ExplanationPostprocessing,
    ExplanationRequest,
    ExplanationResponse,
    ImagePromptItemExplanation,
    ImageScore,
    PromptGranularity,
    TargetGranularity,
    TargetPromptItemExplanation,
    TargetScore,
    TextPromptItemExplanation,
    TextScore,
    TokenPromptItemExplanation,
    TokenScore,
)
from .tokenization import TokenizationRequest, TokenizationResponse
from .reranking import RerankRequest, RerankResponse, RerankResult, RerankUsage
from .utils import load_base64_from_file, load_base64_from_url
from .version import __version__

__all__ = [
    "AsyncClient",
    "BatchSemanticEmbeddingRequest",
    "BatchSemanticEmbeddingResponse",
    "BusyError",
    "ChatRequest",
    "ChatResponse",
    "Client",
    "CompletionRequest",
    "CompletionResponse",
    "ControlTokenOverlap",
    "CustomGranularity",
    "DetokenizationRequest",
    "DetokenizationResponse",
    "Document",
    "EmbeddingRequest",
    "EmbeddingV2Request",
    "EmbeddingResponse",
    "EmbeddingV2Response",
    "EvaluationRequest",
    "EvaluationResponse",
    "Explanation",
    "ExplanationPostprocessing",
    "ExplanationRequest",
    "ExplanationResponse",
    "Image",
    "ImageControl",
    "ImagePromptItemExplanation",
    "ImageScore",
    "load_base64_from_file",
    "load_base64_from_url",
    "Message",
    "POOLING_OPTIONS",
    "Prompt",
    "PromptTemplate",
    "PromptGranularity",
    "QuotaError",
    "RerankRequest",
    "RerankResponse",
    "RerankResult",
    "RerankUsage",
    "SemanticEmbeddingRequest",
    "SemanticEmbeddingResponse",
    "SemanticRepresentation",
    "TargetGranularity",
    "TargetPromptItemExplanation",
    "TargetScore",
    "Text",
    "TextControl",
    "TextPromptItemExplanation",
    "TextScore",
    "TokenControl",
    "TokenizationRequest",
    "TokenizationResponse",
    "TokenPromptItemExplanation",
    "Tokens",
    "TokenScore",
    "__version__",
]
